import ipaddress
import logging
import random
import shutil
import string
from collections import Counter
from unittest.mock import Mock, patch

import pytest
from netsnmpy.netsnmp import SNMPVariable
from netsnmpy.oids import OID as NS_OID
from netsnmpy.trapsession import SNMPTrap

from zino.oid import OID
from zino.trapd.base import (
    TrapObserver,
    TrapOriginator,
)
from zino.trapd.netsnmpy_backend import TrapReceiver

from .. import SNMPTRAP_MISSING
from . import send_trap_externally

OID_COLD_START = ".1.3.6.1.6.3.1.1.5.1"
OID_SYSNAME_0 = ".1.3.6.1.2.1.1.5.0"


class TestTrapReceiver:
    async def test_add_community_should_accept_same_community_multiple_times(self):
        receiver = TrapReceiver()
        receiver.add_community("public")
        receiver.add_community("public")
        assert len(receiver._communities) == 1

    async def test_when_trap_is_well_formed_it_should_be_verified(
        self, trap_message, state_with_localhost, localhost_netsnmpy_receiver
    ):
        device = state_with_localhost.devices.get("localhost")
        origin = TrapOriginator(address=trap_message.source, port=666, device=device)
        assert localhost_netsnmpy_receiver._verify_trap(trap_message, origin)

    async def test_when_trap_lacks_trap_oid_it_should_be_ignored(
        self, trap_message, state_with_localhost, localhost_netsnmpy_receiver
    ):
        device = state_with_localhost.devices.get("localhost")
        origin = TrapOriginator(address=trap_message.source, port=666, device=device)
        trap_message.trap_oid = None
        assert not localhost_netsnmpy_receiver._verify_trap(trap_message, origin)

    async def test_when_trap_lacks_sysuptime_it_should_be_ignored(
        self, trap_message, state_with_localhost, localhost_netsnmpy_receiver
    ):
        device = state_with_localhost.devices.get("localhost")
        origin = TrapOriginator(address=trap_message.source, port=666, device=device)
        trap_message.uptime = None
        assert not localhost_netsnmpy_receiver._verify_trap(trap_message, origin)

    async def test_when_trap_has_incorrect_community_it_should_be_ignored(
        self, trap_message, state_with_localhost, localhost_netsnmpy_receiver
    ):
        device = state_with_localhost.devices.get("localhost")
        origin = TrapOriginator(address=trap_message.source, port=666, device=device)
        trap_message.community = "zaphod"
        assert not localhost_netsnmpy_receiver._verify_trap(trap_message, origin)

    async def test_when_no_community_requirement_is_configured_any_community_should_be_verified(
        self, trap_message, state_with_localhost, localhost_netsnmpy_receiver
    ):
        localhost_netsnmpy_receiver._communities = {}
        device = state_with_localhost.devices.get("localhost")
        origin = TrapOriginator(address=trap_message.source, port=666, device=device)
        trap_message.community = "".join(random.choices(string.ascii_letters + string.digits, k=12))
        assert localhost_netsnmpy_receiver._verify_trap(trap_message, origin)

    async def test_when_trap_observer_wants_no_traps_auto_subscribe_should_ignore_it(self, localhost_netsnmpy_receiver):
        class MockObserver(TrapObserver):
            WANTED_TRAPS = set()

        localhost_netsnmpy_receiver.auto_subscribe_observers()
        assert not any(
            isinstance(observer, MockObserver) for observer in localhost_netsnmpy_receiver._observers.values()
        )

    async def test_when_called_multiple_times_auto_subscribe_should_not_add_duplicates(
        self, localhost_netsnmpy_receiver
    ):
        """The same observer class should not be subscribed more than once for the same trap"""

        class MockObserver(TrapObserver):
            WANTED_TRAPS = {("MOCK-MIB", "mockTrap")}

        localhost_netsnmpy_receiver.auto_subscribe_observers()
        localhost_netsnmpy_receiver.auto_subscribe_observers()

        type_counts = Counter(
            (trap, type(observer))
            for trap, observers in localhost_netsnmpy_receiver._observers.items()
            for observer in observers
        )
        dupes = {ident: count for ident, count in type_counts.items() if count > 1}
        assert not dupes

    async def test_when_trap_is_unknown_it_should_not_crash(self, localhost_netsnmpy_receiver):
        source = ipaddress.ip_address("127.0.0.1")
        trap = SNMPTrap(
            source=source,
            agent=source,
            generic_type=None,
            trap_oid=OID(".1.1.1.1.1.1.1.1"),
            uptime=1984,
            community="public",
            version="2c",
            variables=[
                SNMPVariable(NS_OID(".1.3.6.1.2.1.1.3.0"), 1984),
            ],
        )
        assert not localhost_netsnmpy_receiver.trap_received(trap)

    async def test_when_trap_contains_unknown_variable_it_should_not_crash(self, localhost_netsnmpy_receiver):
        source = ipaddress.ip_address("127.0.0.1")
        trap = SNMPTrap(
            source=source,
            agent=source,
            generic_type=None,
            trap_oid=OID(".1.3.6.1.4.1.2636.5.1.1.1.0.2"),
            uptime=1984,
            community="public",
            version="2c",
            variables=[
                SNMPVariable(NS_OID(".1.3.6.1.2.1.1.3.0"), 1984),
                SNMPVariable(NS_OID(".1.1.1.1.1.1.1.1.1"), 42),
            ],
        )
        assert not localhost_netsnmpy_receiver.trap_received(trap)


@pytest.mark.skipif(not shutil.which("snmptrap"), reason=SNMPTRAP_MISSING)
class TestTrapReceiverExternally:
    async def test_when_trap_is_from_unknown_device_it_should_ignore_it(self, event_loop, unused_udp_port, caplog):
        receiver = TrapReceiver(address="127.0.0.1", port=unused_udp_port, loop=event_loop)
        receiver.add_community("public")
        try:
            await receiver.open()

            with caplog.at_level(logging.DEBUG):
                await send_trap_externally(OID_COLD_START, OID_SYSNAME_0, "s", "'MockDevice'", port=unused_udp_port)
                assert "ignored trap from 127.0.0.1" in caplog.text
        finally:
            receiver.close()

    async def test_when_trap_is_from_known_device_it_should_log_it(self, localhost_netsnmpy_receiver, caplog):
        with caplog.at_level(logging.DEBUG):
            await send_trap_externally(
                OID_COLD_START, OID_SYSNAME_0, "s", "'MockDevice'", port=localhost_netsnmpy_receiver.port
            )
            assert "Trap from localhost" in caplog.text
            assert OID_COLD_START in caplog.text

    async def test_when_observer_is_added_and_trap_matches_it_should_call_it(self, localhost_netsnmpy_receiver):
        observer = Mock()
        localhost_netsnmpy_receiver.observe(observer, ("SNMPv2-MIB", "coldStart"))
        await send_trap_externally(
            OID_COLD_START, OID_SYSNAME_0, "s", "'MockDevice'", port=localhost_netsnmpy_receiver.port
        )
        assert observer.handle_trap.called

    async def test_when_observer_raises_unhandled_exception_it_should_log_it(self, localhost_netsnmpy_receiver, caplog):
        crashing_observer = Mock()
        crashing_observer.handle_trap.side_effect = ValueError("mocked exception")
        localhost_netsnmpy_receiver.observe(crashing_observer, ("SNMPv2-MIB", "coldStart"))

        with caplog.at_level(logging.INFO):
            await send_trap_externally(
                OID_COLD_START, OID_SYSNAME_0, "s", "'MockDevice'", port=localhost_netsnmpy_receiver.port
            )
            assert "ValueError" in caplog.text
            assert "mocked exception" in caplog.text

    async def test_when_early_observer_returns_false_it_should_not_call_later_observers(
        self, localhost_netsnmpy_receiver, event_loop
    ):
        early_observer = Mock()
        false_result = event_loop.create_future()
        false_result.set_result(False)
        early_observer.handle_trap.return_value = false_result
        late_observer = Mock()
        localhost_netsnmpy_receiver.observe(early_observer, ("BGP4-MIB", "bgpBackwardTransition"))
        localhost_netsnmpy_receiver.observe(late_observer, ("BGP4-MIB", "bgpBackwardTransition"))
        bgp_backward_transition_trap = [
            ".1.3.6.1.2.1.15.7.2",
            ".1.3.6.1.2.1.15.3.1.7",
            "a",
            "192.168.42.42",
            ".1.3.6.1.2.1.15.3.1.14",
            "x",
            "4242",
            ".1.3.6.1.2.1.15.3.1.2",
            "i",
            "2",
        ]
        await send_trap_externally(*bgp_backward_transition_trap, port=localhost_netsnmpy_receiver.port)
        assert early_observer.handle_trap.called
        assert not late_observer.handle_trap.called

    async def test_when_trap_verification_fails_it_should_not_dispatch_trap(self, localhost_netsnmpy_receiver):
        with patch.object(localhost_netsnmpy_receiver, "_verify_trap", return_value=False):
            with patch.object(localhost_netsnmpy_receiver, "dispatch_trap") as mock_dispatch:
                await send_trap_externally(
                    OID_COLD_START, OID_SYSNAME_0, "s", "'MockDevice'", port=localhost_netsnmpy_receiver.port
                )
                assert not mock_dispatch.called


@pytest.fixture
def trap_message():
    source = ipaddress.ip_address("127.0.0.1")
    return SNMPTrap(
        source=source,
        agent=source,
        generic_type=None,
        trap_oid=OID(".1.3.6.1.4.1.2636.5.1.1.1.0.2"),
        uptime=1984,
        community="public",
        version="2c",
        variables=[
            SNMPVariable(NS_OID(".1.3.6.1.2.1.1.3.0"), 1984),
            SNMPVariable(NS_OID(".1.1.1.1.1.1.1.1.1"), 42),
        ],
    )



import ipaddress
import logging
from unittest.mock import Mock

import pytest

from zino.statemodels import BGPOperState, BGPPeerSession
from zino.trapd.base import TrapMessage
from zino.trapobservers.bgp_traps import BgpTrapObserver


class TestBgpTrapObserver:
    async def test_when_backward_transition_trap_is_received_it_should_change_bgp_peer_state(
        self, backward_transition_trap
    ):
        device = backward_transition_trap.agent.device
        peer = next(iter(device.bgp_peers.keys()))

        observer = BgpTrapObserver(state=Mock())
        await observer.handle_trap(trap=backward_transition_trap)

        assert len(device.bgp_peers) == 1
        assert device.bgp_peers[peer].oper_state == BGPOperState.ACTIVE

    async def test_when_trap_is_missing_required_varbinds_it_should_do_nothing(self, backward_transition_trap):
        device = backward_transition_trap.agent.device
        peer = next(iter(device.bgp_peers.keys()))
        addr_type = backward_transition_trap.get_all("jnxBgpM2PeerLocalAddrType")[0]
        backward_transition_trap.variables.remove(addr_type)

        observer = BgpTrapObserver(state=Mock())
        await observer.handle_trap(trap=backward_transition_trap)

        assert len(device.bgp_peers) == 1
        assert device.bgp_peers[peer].oper_state == BGPOperState.ESTABLISHED

    async def test_when_trap_has_invalid_remote_addr_it_should_do_nothing(self, backward_transition_trap):
        device = backward_transition_trap.agent.device
        peer = next(iter(device.bgp_peers.keys()))
        backward_transition_trap.get_all("jnxBgpM2PeerRemoteAddr")[0].raw_value = b"INVALID"

        observer = BgpTrapObserver(state=Mock())
        await observer.handle_trap(trap=backward_transition_trap)

        assert len(device.bgp_peers) == 1
        assert device.bgp_peers[peer].oper_state == BGPOperState.ESTABLISHED

    async def test_when_trap_has_invalid_oper_state_it_should_do_nothing(self, backward_transition_trap):
        device = backward_transition_trap.agent.device
        peer = next(iter(device.bgp_peers.keys()))
        backward_transition_trap.get_all("jnxBgpM2PeerState")[0].value = "INVALIDFOOBAR"

        observer = BgpTrapObserver(state=Mock())
        await observer.handle_trap(trap=backward_transition_trap)

        assert len(device.bgp_peers) == 1
        assert device.bgp_peers[peer].oper_state == BGPOperState.ESTABLISHED

    async def test_when_established_trap_is_received_it_should_just_log_it(self, established_trap, caplog):
        """This requirement is disputed until Håvard E confirms it"""
        observer = BgpTrapObserver(state=Mock())
        with caplog.at_level(logging.INFO):
            await observer.handle_trap(trap=established_trap)
            assert "BGP peer up" in caplog.text

    async def test_when_trap_is_unknown_it_should_pass_it_on(self, established_trap):
        established_trap.name = "FOOBAR"
        observer = BgpTrapObserver(state=Mock())
        assert await observer.handle_trap(trap=established_trap)


@pytest.fixture
def backward_transition_trap(localhost_trap_originator) -> TrapMessage:
    """Returns a correct backward transition trap with internal state to match"""
    peer = ipaddress.IPv4Address("10.0.0.1")
    localhost_trap_originator.device.bgp_peers = {peer: BGPPeerSession(oper_state=BGPOperState.ESTABLISHED)}

    trap = TrapMessage(agent=localhost_trap_originator, mib="BGP4-V2-MIB-JUNIPER", name="jnxBgpM2BackwardTransition")
    trap.variables = [
        Mock(var="jnxBgpM2PeerLocalAddrType", value=1),
        Mock(var="jnxBgpM2PeerRemoteAddrType", value=1),
        Mock(var="jnxBgpM2PeerRemoteAddr", raw_value=peer.packed),
        Mock(var="jnxBgpM2PeerState", value="active"),
    ]
    return trap


@pytest.fixture
def established_trap(localhost_trap_originator) -> TrapMessage:
    """Returns a correct established trap with internal state to match"""
    peer = ipaddress.IPv4Address("10.0.0.1")
    localhost_trap_originator.device.bgp_peers = {peer: BGPPeerSession(oper_state=BGPOperState.ACTIVE)}

    trap = TrapMessage(agent=localhost_trap_originator, mib="BGP4-V2-MIB-JUNIPER", name="jnxBgpM2Established")
    trap.variables = [
        Mock(var="jnxBgpM2PeerLocalAddrType", value=1),
        Mock(var="jnxBgpM2PeerRemoteAddrType", value=1),
        Mock(var="jnxBgpM2PeerRemoteAddr", raw_value=peer.packed),
        Mock(var="jnxBgpM2PeerState", value="established"),
    ]
    return trap



#
# PySNMP MIB module OLD-CISCO-TCP-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file:///home/mvold/mibs/cisco/OLD-CISCO-TCP-MIB.my
# Produced by pysmi-0.3.4 at Thu Jul 18 11:03:56 2024
# On host agrajag platform Linux version 6.6.37 by user mvold
# Using Python version 3.11.9 (main, Apr  2 2024, 08:25:04) [GCC 13.2.0]
#
Integer, ObjectIdentifier, OctetString = mibBuilder.importSymbols("ASN1", "Integer", "ObjectIdentifier", "OctetString")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
SingleValueConstraint, ValueSizeConstraint, ValueRangeConstraint, ConstraintsUnion, ConstraintsIntersection = mibBuilder.importSymbols("ASN1-REFINEMENT", "SingleValueConstraint", "ValueSizeConstraint", "ValueRangeConstraint", "ConstraintsUnion", "ConstraintsIntersection")
local, = mibBuilder.importSymbols("CISCO-SMI", "local")
NotificationGroup, ModuleCompliance = mibBuilder.importSymbols("SNMPv2-CONF", "NotificationGroup", "ModuleCompliance")
Bits, Counter64, IpAddress, iso, Gauge32, Integer32, ModuleIdentity, Unsigned32, MibIdentifier, ObjectIdentity, MibScalar, MibTable, MibTableRow, MibTableColumn, NotificationType, Counter32, TimeTicks = mibBuilder.importSymbols("SNMPv2-SMI", "Bits", "Counter64", "IpAddress", "iso", "Gauge32", "Integer32", "ModuleIdentity", "Unsigned32", "MibIdentifier", "ObjectIdentity", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "NotificationType", "Counter32", "TimeTicks")
TextualConvention, DisplayString = mibBuilder.importSymbols("SNMPv2-TC", "TextualConvention", "DisplayString")
tcpConnRemPort, tcpConnLocalPort, tcpConnLocalAddress, tcpConnRemAddress = mibBuilder.importSymbols("TCP-MIB", "tcpConnRemPort", "tcpConnLocalPort", "tcpConnLocalAddress", "tcpConnRemAddress")
ltcp = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 2, 6))
ltcpConnTable = MibTable((1, 3, 6, 1, 4, 1, 9, 2, 6, 1), )
if mibBuilder.loadTexts: ltcpConnTable.setStatus('deprecated')
ltcpConnEntry = MibTableRow((1, 3, 6, 1, 4, 1, 9, 2, 6, 1, 1), ).setIndexNames((0, "TCP-MIB", "tcpConnLocalAddress"), (0, "TCP-MIB", "tcpConnLocalPort"), (0, "TCP-MIB", "tcpConnRemAddress"), (0, "TCP-MIB", "tcpConnRemPort"))
if mibBuilder.loadTexts: ltcpConnEntry.setStatus('deprecated')
loctcpConnInBytes = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 2, 6, 1, 1, 1), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: loctcpConnInBytes.setStatus('deprecated')
loctcpConnOutBytes = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 2, 6, 1, 1, 2), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: loctcpConnOutBytes.setStatus('deprecated')
loctcpConnInPkts = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 2, 6, 1, 1, 3), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: loctcpConnInPkts.setStatus('deprecated')
loctcpConnOutPkts = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 2, 6, 1, 1, 4), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: loctcpConnOutPkts.setStatus('deprecated')
loctcpConnElapsed = MibTableColumn((1, 3, 6, 1, 4, 1, 9, 2, 6, 1, 1, 5), TimeTicks()).setMaxAccess("readonly")
if mibBuilder.loadTexts: loctcpConnElapsed.setStatus('deprecated')
mibBuilder.exportSymbols("OLD-CISCO-TCP-MIB", loctcpConnOutBytes=loctcpConnOutBytes, loctcpConnOutPkts=loctcpConnOutPkts, loctcpConnInBytes=loctcpConnInBytes, loctcpConnElapsed=loctcpConnElapsed, ltcp=ltcp, ltcpConnEntry=ltcpConnEntry, loctcpConnInPkts=loctcpConnInPkts, ltcpConnTable=ltcpConnTable)



"""This module implements BGP trap handling.

Examples of how to send test traps:

snmptrap -v 2c -c public localhost:1162 "" \
    BGP4-V2-MIB-JUNIPER::jnxBgpM2BackwardTransition \
    BGP4-V2-MIB-JUNIPER::jnxBgpM2PeerLocalAddrType i 1 \
    BGP4-V2-MIB-JUNIPER::jnxBgpM2PeerLocalAddr x "0A000002" \
    BGP4-V2-MIB-JUNIPER::jnxBgpM2PeerRemoteAddrType i 1 \
    BGP4-V2-MIB-JUNIPER::jnxBgpM2PeerRemoteAddr x "0A000001" \
    BGP4-V2-MIB-JUNIPER::jnxBgpM2PeerLastErrorReceived x "0102" \
    BGP4-V2-MIB-JUNIPER::jnxBgpM2PeerState i 4

This sends all the variables required by the MIB, but this trap observer only cares about the remote peer address and
the peer state value.

"""

import logging
from ipaddress import ip_address
from typing import Optional, Tuple

from zino.statemodels import BGPOperState, BGPPeerSession, IPAddress
from zino.trapd.base import TrapMessage, TrapObserver

_logger = logging.getLogger(__name__)


class BgpTrapObserver(TrapObserver):
    """Handles BGP peering session operational transition messages"""

    WANTED_TRAPS = {
        ("BGP4-V2-MIB-JUNIPER", "jnxBgpM2BackwardTransition"),
        ("BGP4-V2-MIB-JUNIPER", "jnxBgpM2Established"),
    }

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        try:
            peer, state = self._pre_parse_trap(trap)
        except MissingRequiredTrapVariables:
            return
        except ValueError as error:
            _logger.warning(error)
            return

        if trap.name == "jnxBgpM2BackwardTransition":
            self.handle_backward_transition(trap, peer, state)
        elif trap.name == "jnxBgpM2Established":
            self.handle_established(trap, peer, state)
        else:
            # Something weird happened, let someone else handle it
            _logger.info("%s: Unknown trap received: %s", trap.agent.device.name, trap.name)
            return True

    def handle_backward_transition(self, trap: TrapMessage, peer: IPAddress, state: BGPOperState):
        _logger.debug("BGP backward transition trap received: %r", trap)
        bgp_peers = trap.agent.device.bgp_peers
        prev_state = bgp_peers[peer].oper_state if peer in bgp_peers else "unknown"

        if state != BGPOperState.ESTABLISHED and prev_state == BGPOperState.ESTABLISHED:
            _logger.info("%s Lost BGP peer: %s state %s", trap.agent.device.name, peer, state)

        bgp_peers.setdefault(peer, BGPPeerSession()).oper_state = state

    def handle_established(self, trap: TrapMessage, peer: IPAddress, state: BGPOperState):
        _logger.debug("BGP established trap received: %r", trap)
        # TODO Zino 1 does not actually update the internal peering state here, we should verify that this is really
        #  the desired behavior
        _logger.info("%s BGP peer up: %s state %s", trap.agent.device.name, peer, state)

    def _pre_parse_trap(self, trap: TrapMessage) -> Tuple[IPAddress, BGPOperState]:
        if "jnxBgpM2PeerLocalAddrType" not in trap:
            raise MissingRequiredTrapVariables()

        try:
            remote_addr = bytes(trap.get_all("jnxBgpM2PeerRemoteAddr")[0].raw_value)
            peer = ip_address(remote_addr)
        except ValueError:
            raise ValueError(f"BGP transition trap received with invalid peer address: {remote_addr!r}")
        except IndexError:
            raise ValueError("BGP transition trap received without peer address")

        try:
            raw_state = trap.get_all("jnxBgpM2PeerState")[0].value
            state = BGPOperState(raw_state)
        except ValueError:
            raise ValueError(f"BGP transition trap received with invalid peer state: {raw_state}")
        except IndexError:
            raise ValueError("BGP transition trap received without peer state")

        return peer, state


class MissingRequiredTrapVariables(ValueError):
    pass



from unittest.mock import Mock
from weakref import WeakValueDictionary

import pytest

import zino.snmp
from zino.config.models import PollDevice
from zino.snmp import get_snmp_session, import_snmp_backend


class TestGetSnmpSession:
    def test_it_should_return_same_session_instance_for_same_device(self, monkeypatch):
        # Temporarily enable session re-use (which is otherwise disabled for testing purposes)
        with monkeypatch.context() as patch:
            patch.setattr("zino.snmp._snmp_sessions", WeakValueDictionary())
            device = PollDevice(name="localhost", address="127.0.0.1", community="public", snmpversion="v2c")
            session1 = get_snmp_session(device)
            session2 = get_snmp_session(device)
            assert session1 is session2

    def test_when_reuse_is_disabled_it_should_return_new_session(self):
        device = PollDevice(name="localhost", address="127.0.0.1", community="public", snmpversion="v2c")
        session1 = get_snmp_session(device)
        session2 = get_snmp_session(device)
        assert session1 is not session2

    def test_when_backend_is_not_loaded_it_should_raise(self, monkeypatch):
        with monkeypatch.context() as patch:
            patch.setattr("zino.snmp._selected_backend", None)
            with pytest.raises(zino.snmp.SNMPBackendNotLoaded):
                get_snmp_session(Mock())


class TestImportSnmpBackend:
    def test_when_backend_is_already_selected_and_argument_is_empty_it_should_return_already_loaded_backend(self):
        assert zino.snmp._selected_backend is not None  # conftest should already have loaded a backend
        module = import_snmp_backend()
        assert module == zino.snmp._selected_backend

    def test_when_backend_is_unknown_it_should_raise_value_error(self):
        with pytest.raises(ValueError):
            import_snmp_backend("nonexistent-foobar")



import logging
from datetime import datetime, timezone

from zino.state import ZinoState
from zino.stateconverter.linedata import LineData
from zino.stateconverter.utils import OldState, parse_log_and_history
from zino.statemodels import (
    DeviceMaintenance,
    MatchType,
    PlannedMaintenance,
    PmType,
    PortStateMaintenance,
)

_log = logging.getLogger(__name__)


PM_CLASS_MAPPING = {
    PmType.PORTSTATE: PortStateMaintenance,
    PmType.DEVICE: DeviceMaintenance,
}


def set_pm_state(old_state: OldState, new_state: ZinoState):
    _set_pm_events(old_state, new_state)
    for linedata in old_state["::pm::lastid"]:
        _set_last_id(linedata, new_state)
    for linedata in old_state["::pm::lasttime"]:
        _set_last_time(linedata, new_state)


def _set_last_id(linedata: LineData, state: ZinoState):
    """Sets the last planned maintenance id"""
    state.planned_maintenances.last_pm_id = int(linedata.value)


def _set_last_time(linedata: LineData, state: ZinoState):
    """Sets timestamp for last time planned maintenance was run"""
    timestamp = int(linedata.value)
    state.planned_maintenances.last_run = datetime.fromtimestamp(timestamp, tz=timezone.utc)


def _set_pm_events(old_state: OldState, new_state: ZinoState):
    """Registers PlannedMaintenance objects"""
    pm_data = _group_pm_attrs_by_id(old_state)
    event_mapping = _get_events_affected_by_pms(old_state)
    for id, attrs in pm_data.items():
        try:
            pm = _create_pm(attrs)
        except ValueError as e:
            _log.error(f"Error setting pm_event attribute: {str(e)}")
        pm.event_ids = event_mapping.get(id, [])
        pm.id = id
        new_state.planned_maintenances.planned_maintenances[id] = pm


def _create_pm(attrs: dict[str, str]) -> PlannedMaintenance:
    """Creates a PlannedMaintenance object from attributes"""

    required_attrs = ["starttime", "endtime", "match_type", "match_expr", "type"]
    for required_attr in required_attrs:
        if required_attr not in attrs:
            raise ValueError(f"Required attribute '{required_attr}' not found in pm_event")

    start_time = datetime.fromtimestamp(int(attrs["starttime"]), tz=timezone.utc)
    end_time = datetime.fromtimestamp(int(attrs["endtime"]), tz=timezone.utc)
    match_type = MatchType(attrs["match_type"])
    match_expression = attrs["match_expr"]
    match_device = attrs.get("match_dev")
    pm_type = PmType(attrs["type"])
    log = parse_log_and_history(attrs.get("log", ""))

    pm_class = PM_CLASS_MAPPING.get(pm_type)
    if pm_class is None:
        raise ValueError(f"Unknown pm_event type {pm_type}")

    pm = pm_class(
        start_time=start_time,
        end_time=end_time,
        match_type=match_type,
        match_expression=match_expression,
        match_device=match_device,
        log=log,
    )

    return pm


def _group_pm_attrs_by_id(old_state: OldState) -> dict[int, dict[str, str]]:
    """Returns a dict of dicts with mapping from pm_event ID to pm_event attributes"""
    return_dict = dict()
    for linedata in old_state["::pm::event"]:
        pm_attr = linedata.identifiers[0]
        pm_id = int(linedata.identifiers[1])
        if pm_id not in return_dict:
            return_dict[pm_id] = dict()
        return_dict[pm_id][pm_attr] = linedata.value
    return return_dict


def _get_events_affected_by_pms(old_state: OldState) -> dict[int, list[int]]:
    """Returns a mapping of pm ID to list of event IDs affected by that PM"""
    pm_to_events = dict()
    for linedata in old_state["::pm::pm_events"]:
        pm_id = pm_id = int(linedata.identifiers[0])
        event_ids = [int(id) for id in linedata.value.split()]
        pm_to_events[pm_id] = event_ids
    return pm_to_events



"""Simple tests for the getuptime example program"""

import subprocess

import pytest


def test_get_uptime_should_run_without_error(polldevs_with_localhost, snmpsim):
    confdir = polldevs_with_localhost.parent
    assert subprocess.check_call(["python3", "-m", "zino.getuptime", "localhost"], cwd=confdir) == 0


@pytest.fixture
def polldevs_with_localhost(zino_conf, snmp_test_port):
    polldevs_conf = zino_conf.parent.joinpath("polldevs.cf")
    with open(polldevs_conf, "w") as conf:
        conf.write(
            f"""# polldevs test config
            default interval: 10
            default community: public
            default domain: uninett.no
            default statistics: yes
            default snmpversion: v2c

            name: localhost
            address: 127.0.0.1
            port: {snmp_test_port}
            """
        )
    yield polldevs_conf



import logging

from zino.debug import debug_log_timeout_error
from zino.snmp import get_snmp_session
from zino.tasks.errors import DeviceUnreachableError

_log = logging.getLogger(__name__)


async def run_all_tasks(device, state, config):
    try:
        with get_snmp_session(device):
            await run_registered_tasks(device, state, config)
    except DeviceUnreachableError:
        _log.debug(f"Device {device.name} could not be reached. Any remaining tasks have been cancelled.")


async def run_registered_tasks(device, state, config):
    for task_class in get_registered_tasks():
        task = task_class(device, state, config)
        try:
            await task.run()
        except TimeoutError:
            _log.error(
                "%s: %s raised an unexpected TimeoutError mid-run, cancelling remaining tasks in this run",
                device.name,
                task_class.__name__,
            )
            debug_log_timeout_error(device.name, task_class, logger=_log)
            return


def get_registered_tasks():
    from zino.tasks.addrs import AddressMapTask
    from zino.tasks.bfdtask import BFDTask
    from zino.tasks.bgpstatemonitortask import BGPStateMonitorTask
    from zino.tasks.juniperalarmtask import JuniperAlarmTask
    from zino.tasks.linkstatetask import LinkStateTask
    from zino.tasks.reachabletask import ReachableTask
    from zino.tasks.vendor import VendorTask

    return [ReachableTask, VendorTask, AddressMapTask, LinkStateTask, BFDTask, BGPStateMonitorTask, JuniperAlarmTask]



#
# PySNMP MIB module DIFFSERV-DSCP-TC (http://snmplabs.com/pysmi)
# ASN.1 source file:///usr/share/snmp/mibs/ietf/DIFFSERV-DSCP-TC
# Produced by pysmi-1.1.10 at Mon Oct  9 12:14:08 2023
# On host simtve-ThinkPad-X1-Carbon-Gen-9 platform Linux version 5.15.0-84-generic by user simon
# Using Python version 3.9.18 (main, Aug 25 2023, 13:20:04) 
#
Integer, ObjectIdentifier, OctetString = mibBuilder.importSymbols("ASN1", "Integer", "ObjectIdentifier", "OctetString")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsUnion, ValueRangeConstraint, ValueSizeConstraint, ConstraintsIntersection, SingleValueConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsUnion", "ValueRangeConstraint", "ValueSizeConstraint", "ConstraintsIntersection", "SingleValueConstraint")
NotificationGroup, ModuleCompliance = mibBuilder.importSymbols("SNMPv2-CONF", "NotificationGroup", "ModuleCompliance")
IpAddress, ModuleIdentity, Bits, MibIdentifier, Counter64, Integer32, iso, TimeTicks, NotificationType, Unsigned32, mib_2, MibScalar, MibTable, MibTableRow, MibTableColumn, ObjectIdentity, Gauge32, Counter32 = mibBuilder.importSymbols("SNMPv2-SMI", "IpAddress", "ModuleIdentity", "Bits", "MibIdentifier", "Counter64", "Integer32", "iso", "TimeTicks", "NotificationType", "Unsigned32", "mib-2", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "ObjectIdentity", "Gauge32", "Counter32")
DisplayString, TextualConvention = mibBuilder.importSymbols("SNMPv2-TC", "DisplayString", "TextualConvention")
diffServDSCPTC = ModuleIdentity((1, 3, 6, 1, 2, 1, 96))
diffServDSCPTC.setRevisions(('2002-05-09 00:00',))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    if mibBuilder.loadTexts: diffServDSCPTC.setRevisionsDescriptions(('Initial version, published as RFC 3289.',))
if mibBuilder.loadTexts: diffServDSCPTC.setLastUpdated('200205090000Z')
if mibBuilder.loadTexts: diffServDSCPTC.setOrganization('IETF Differentiated Services WG')
if mibBuilder.loadTexts: diffServDSCPTC.setContactInfo(' Fred Baker Cisco Systems 1121 Via Del Rey Santa Barbara, CA 93117, USA E-mail: fred@cisco.com Kwok Ho Chan Nortel Networks 600 Technology Park Drive Billerica, MA 01821, USA E-mail: khchan@nortelnetworks.com Andrew Smith Harbour Networks Jiuling Building 21 North Xisanhuan Ave. Beijing, 100089, PRC E-mail: ah_smith@acm.org Differentiated Services Working Group: diffserv@ietf.org')
if mibBuilder.loadTexts: diffServDSCPTC.setDescription('The Textual Conventions defined in this module should be used whenever a Differentiated Services Code Point is used in a MIB.')
class Dscp(TextualConvention, Integer32):
    reference = 'RFC 2474, RFC 2780'
    description = 'A Differentiated Services Code-Point that may be used for marking a traffic stream.'
    status = 'current'
    displayHint = 'd'
    subtypeSpec = Integer32.subtypeSpec + ValueRangeConstraint(0, 63)

class DscpOrAny(TextualConvention, Integer32):
    reference = 'RFC 2474, RFC 2780'
    description = 'The IP header Differentiated Services Code-Point that may be used for discriminating among traffic streams. The value -1 is used to indicate a wild card i.e. any value.'
    status = 'current'
    displayHint = 'd'
    subtypeSpec = Integer32.subtypeSpec + ConstraintsUnion(ValueRangeConstraint(-1, -1), ValueRangeConstraint(0, 63), )
mibBuilder.exportSymbols("DIFFSERV-DSCP-TC", DscpOrAny=DscpOrAny, Dscp=Dscp, PYSNMP_MODULE_ID=diffServDSCPTC, diffServDSCPTC=diffServDSCPTC)



# mypy: ignore-errors
"""Compatibility layer for older versions of Python"""

try:
    from enum import StrEnum
except ImportError:
    # < Python 3.11
    from enum import Enum

    class StrEnum(str, Enum):
        __str__ = str.__str__

        def _generate_next_value_(name, start, count, last_values):
            """
            Return the lower-cased version of the member name.
            """
            return name.lower()


__all__ = ["StrEnum"]



"""Base SNMP functionality and types shared by all back-ends"""

from dataclasses import dataclass
from typing import Any, NamedTuple, Union

from zino.oid import OID


@dataclass
class MibObject:
    oid: OID
    value: Union[str, int, OID]


class Identifier(NamedTuple):
    """Identifies a MIB object by MIB, object name and row index"""

    mib: str
    object: str
    index: OID


SNMPVarBind = tuple[Identifier, Any]
SparseWalkResponse = dict[OID, dict[str, Any]]


class SnmpError(Exception):
    """Base class for SNMP, MIB and OID specific errors"""


class ErrorIndication(SnmpError):
    """Class for SNMP errors that occur locally,
    as opposed to being reported from a different SNMP entity.
    """


class MibNotFoundError(ErrorIndication):
    """Raised if a required MIB file could not be found"""


class ErrorStatus(SnmpError):
    """Raised if an SNMP entity includes a non-zero error status in its response PDU.
    RFC 1905 defines the possible errors that can be specified in the error status field.
    This can either be used directly or subclassed for one of these specific errors.
    """


class NoSuchNameError(ErrorStatus):
    """Represents the "noSuchName" error. Raised if an object could not be found at an OID."""


class VarBindError(SnmpError):
    """Base class for errors carried in varbinds and not in the errorStatus or errorIndication fields"""


class NoSuchObjectError(VarBindError):
    """Raised if an object could not be found at an OID"""


class NoSuchInstanceError(VarBindError):
    """Raised if an instance could not be found at an OID"""


class EndOfMibViewError(VarBindError):
    """Raised if end of MIB view is encountered"""


class SNMPBackendError(SnmpError):
    """Base class for SNMP back-end specific errors"""


class SNMPBackendNotLoaded(SNMPBackendError):
    """No SNMP back-end has been loaded yet"""


class SNMPBackendVersionError(SNMPBackendError):
    """The available SNMP back-end library version is incompatible with the required version"""



"""Zino configuration models"""

from ipaddress import IPv4Address, IPv6Address
from os import R_OK, access
from os.path import isfile
from typing import Any, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

DEFAULT_INTERVAL_MINUTES = 5
STATE_FILENAME = "zino-state.json"
EVENT_DUMP_DIR = "old-events"
POLLFILE = "polldevs.cf"

IPAddress = Union[IPv4Address, IPv6Address]


def validate_file_can_be_opened(filename: str) -> str:
    assert isfile(filename) and access(filename, R_OK), f"File {filename} doesn't exist or isn't readable"
    return filename


ExistingFileName = Annotated[str, AfterValidator(validate_file_can_be_opened)]


# config fields and default values from
# https://gitlab.sikt.no/verktoy/zino/blob/master/common/config.tcl#L18-44
class PollDevice(BaseModel):
    """Defines the attributes Zino needs/wants to be able to poll a device"""

    name: str
    address: IPAddress
    community: str = "public"
    dns: str = None
    interval: int = DEFAULT_INTERVAL_MINUTES
    ignorepat: Optional[str] = None
    watchpat: Optional[str] = None
    priority: int = 100
    timeout: int = 5
    retries: int = 3
    domain: str = None
    statistics: bool = True
    hcounters: bool = True  # Not used but kept for compatibility with old config files
    snmpversion: Literal["v1", "v2c"] = "v2c"
    do_bgp: bool = True
    port: int = 161


class Archiving(BaseModel):
    model_config = ConfigDict(extra="forbid")

    old_events_dir: str = EVENT_DUMP_DIR


class Authentication(BaseModel):
    model_config = ConfigDict(extra="forbid")

    file: ExistingFileName = "secrets"


class Persistence(BaseModel):
    model_config = ConfigDict(extra="forbid")

    file: str = STATE_FILENAME
    period: int = DEFAULT_INTERVAL_MINUTES


class Polling(BaseModel):
    model_config = ConfigDict(extra="forbid")

    file: ExistingFileName = POLLFILE
    period: int = 1


class TrapConfiguration(BaseModel):
    """Trap reception configuration"""

    model_config = ConfigDict(extra="forbid")

    require_community: list[str] = []


class AgentConfiguration(BaseModel):
    """SNMP agent configuration for responding to uptime queries"""

    model_config = ConfigDict(extra="forbid")

    enabled: bool = True
    port: int = 8000
    address: str = "0.0.0.0"
    community: Optional[str] = "public"


class SNMPConfiguration(BaseModel):
    model_config = ConfigDict(extra="forbid")

    backend: Literal["pysnmp", "netsnmp"] = "netsnmp"
    trap: TrapConfiguration = TrapConfiguration()
    agent: AgentConfiguration = AgentConfiguration()


class SchedulerConfiguration(BaseModel):
    """Options to control the job scheduler behavior"""

    model_config = ConfigDict(extra="forbid")

    misfire_grace_time: int = 10  # Seconds


class EventConfiguration(BaseModel):
    """Options to control how events are created"""

    model_config = ConfigDict(extra="forbid")

    make_events_for_new_interfaces: bool = False
    make_events_for_new_devices: bool = False


class Configuration(BaseModel):
    """Class for keeping track of the configuration set by zino.toml"""

    # throw ValidationError on extra keys
    model_config = ConfigDict(extra="forbid")

    archiving: Archiving = Archiving()
    authentication: Authentication = Authentication()
    persistence: Persistence = Persistence()
    polling: Polling = Polling()
    snmp: SNMPConfiguration = SNMPConfiguration()
    scheduler: SchedulerConfiguration = SchedulerConfiguration()
    event: EventConfiguration = EventConfiguration()
    logging: dict[str, Any] = {
        "version": 1,
        "disable_existing_loggers": False,
        "loggers": {
            "root": {"level": "INFO", "handlers": ["console"]},
            "apscheduler": {"level": "WARNING"},
        },
        "formatters": {"standard": {"format": "%(asctime)s - %(levelname)s - %(name)s (%(threadName)s) - %(message)s"}},
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "stream": "ext://sys.stderr",
            }
        },
    }



"""Tests for zino.debug module."""

import logging
from unittest.mock import AsyncMock, Mock, patch

import pytest

from zino.config.models import PollDevice
from zino.debug import debug_log_timeout_error
from zino.state import ZinoState
from zino.tasks.task import Task


class DummyTask(Task):
    """Dummy task for testing."""

    async def run(self):
        """Run the task."""
        await self.do_snmp_operation()

    async def do_snmp_operation(self):
        """Simulate an SNMP operation that times out."""
        pass


class TestDebugLogTimeoutError:
    """Tests for debug_log_timeout_error function."""

    @pytest.fixture
    def dummy_device(self):
        """Create a dummy device."""
        return PollDevice(
            name="test-router",
            address="10.0.0.1",
            community="public",
        )

    @pytest.fixture
    def dummy_task(self, dummy_device):
        """Create a dummy task instance."""
        return DummyTask(dummy_device, ZinoState(), Mock())

    async def test_it_should_log_task_details(self, dummy_task, caplog):
        """Test that task frame information is logged correctly."""

        # Make the task's do_snmp_operation raise a TimeoutError
        async def raise_timeout():
            raise TimeoutError("Simulated timeout")

        dummy_task.do_snmp_operation = raise_timeout

        try:
            await dummy_task.run()
        except TimeoutError:
            with caplog.at_level(logging.DEBUG):
                debug_log_timeout_error("test-router", DummyTask)

        assert "test-router: TimeoutError in DummyTask" in caplog.text
        assert "test_debug.py" in caplog.text

    async def test_when_called_without_task_frame_in_stack_then_it_should_not_crash(self, caplog):
        """Test behavior when no task frame is found in stack."""
        try:
            raise TimeoutError("Direct timeout")
        except TimeoutError:
            with caplog.at_level(logging.DEBUG):
                debug_log_timeout_error("test-router", Mock)

        assert "TimeoutError in" not in caplog.text

    async def test_when_called_without_snmp_frame_in_stack_then_it_should_only_log_task_info(self, dummy_task, caplog):
        """Test behavior when no SNMP frame is found in stack."""

        # Create a timeout that happens in the task but not in SNMP
        async def non_snmp_timeout():
            raise TimeoutError("Non-SNMP timeout")

        dummy_task.do_snmp_operation = non_snmp_timeout

        try:
            await dummy_task.run()
        except TimeoutError:
            with caplog.at_level(logging.DEBUG):
                debug_log_timeout_error("test-router", DummyTask)

        assert "test-router: TimeoutError in DummyTask" in caplog.text
        assert "SNMP operation failed" not in caplog.text

    async def test_when_real_snmp_object_times_out_then_it_should_be_detected(self, dummy_device, caplog):
        """Test with a more realistic SNMP timeout scenario."""
        from zino.snmp import SNMP

        class RealSnmpTask(Task):
            async def run(self):
                result = await self.snmp.get("SNMPv2-MIB", "sysUpTime", 0)
                return result

        task = RealSnmpTask(dummy_device, ZinoState(), Mock())

        mock_snmp_obj = Mock(spec=SNMP)
        mock_snmp_obj.__class__.__name__ = "SNMP"

        with patch.object(task, "snmp", mock_snmp_obj):
            task.snmp.get = AsyncMock(side_effect=TimeoutError())

            try:
                await task.run()
            except TimeoutError:
                with caplog.at_level(logging.DEBUG):
                    debug_log_timeout_error("test-router", RealSnmpTask)

        assert "test-router: TimeoutError in RealSnmpTask" in caplog.text

    async def test_when_timeout_occurs_in_snmp_operation_then_both_task_and_snmp_details_should_be_logged(
        self, dummy_task, caplog
    ):
        """Test that both task frame and SNMP operation details are logged."""
        import collections
        from unittest.mock import MagicMock

        with patch("zino.debug.inspect.trace") as mock_trace:
            # Create mock frames for both task and SNMP
            task_frame = Mock()
            task_frame.f_locals = {"self": dummy_task}

            snmp_frame = Mock()
            snmp_obj = MagicMock()
            snmp_obj.__class__.__name__ = "SNMP"
            snmp_frame.f_locals = {"self": snmp_obj, "oid": ("IF-MIB", "ifName"), "max_repetitions": 10}

            # Create FrameInfo objects
            FrameInfo = collections.namedtuple(
                "FrameInfo", ["frame", "filename", "lineno", "function", "code_context", "index"]
            )
            task_frame_info = FrameInfo(task_frame, "tasks.py", 50, "run", ["await self.snmp.getbulk(...)"], 0)
            snmp_frame_info = FrameInfo(
                snmp_frame, "snmp.py", 100, "getbulk", ["return await self.session.agetbulk(...)"], 0
            )

            mock_trace.return_value = [snmp_frame_info, task_frame_info]

            with caplog.at_level(logging.DEBUG):
                debug_log_timeout_error("test-router", DummyTask)

        # Should log both task info and SNMP details
        assert "test-router: TimeoutError in DummyTask.run" in caplog.text
        assert "test-router: SNMP operation failed: operation=GET-BULK" in caplog.text
        assert "max_repetitions=10" in caplog.text


async def test_extract_snmp_operation_details_should_use_first_snmp_frame():
    """Test that _extract_snmp_operation_details returns first SNMP frame found."""
    from unittest.mock import MagicMock

    from zino.debug import _extract_snmp_operation_details

    # Create mock frames
    frame1 = MagicMock()
    frame1.frame.f_locals = {"self": MagicMock(__class__=type("NotSNMP", (), {})), "oid": "test1"}
    frame1.function = "some_function"

    frame2 = MagicMock()
    snmp_obj = MagicMock()
    snmp_obj.__class__.__name__ = "SNMP"
    frame2.frame.f_locals = {"self": snmp_obj, "oid": "test2", "max_repetitions": 10}
    frame2.function = "getbulk"

    frame3 = MagicMock()
    frame3.frame.f_locals = {"self": snmp_obj, "oid": "test3"}
    frame3.function = "get"

    frames = [frame1, frame2, frame3]

    result = _extract_snmp_operation_details(frames)

    # Should return details from frame2 (first SNMP frame)
    assert result == "operation=GET-BULK, oid=test2, max_repetitions=10"


@pytest.mark.parametrize(
    "function_name,locals_dict,expected_result",
    [
        ("get", {"oid": ("SNMPv2-MIB", "sysUpTime", 0)}, "operation=GET, oid=('SNMPv2-MIB', 'sysUpTime', 0)"),
        (
            "get2",
            {"variables": [("SNMPv2-MIB", "sysUpTime", 0)]},
            "operation=GET, variables=[('SNMPv2-MIB', 'sysUpTime', 0)]",
        ),
        ("getnext", {"oid": ("IF-MIB", "ifName", "1")}, "operation=GET-NEXT, oid=('IF-MIB', 'ifName', '1')"),
        ("getnext2", {"variables": [("IF-MIB", "ifName")]}, "operation=GET-NEXT, variables=[('IF-MIB', 'ifName')]"),
        ("walk", {"oid": ("IF-MIB", "ifName")}, "operation=WALK (via GET-NEXT), oid=('IF-MIB', 'ifName')"),
        (
            "getbulk",
            {"oid": ("IF-MIB", "ifName"), "max_repetitions": 10},
            "operation=GET-BULK, oid=('IF-MIB', 'ifName'), max_repetitions=10",
        ),
        (
            "getbulk2",
            {"variables": [("IF-MIB", "ifName")], "max_repetitions": 15},
            "operation=GET-BULK, variables=[('IF-MIB', 'ifName')], max_repetitions=15",
        ),
        (
            "bulkwalk",
            {"oid": ("IF-MIB", "ifName"), "max_repetitions": 20},
            "operation=GET-BULK, oid=('IF-MIB', 'ifName'), max_repetitions=20",
        ),
        (
            "sparsewalk",
            {"variables": [("IF-MIB", "ifName")], "max_repetitions": 25},
            "operation=GET-BULK, variables=[('IF-MIB', 'ifName')], max_repetitions=25",
        ),
        ("custom_op", {"oid": "some_oid"}, "operation=custom_op, oid=some_oid"),
    ],
)
async def test_format_snmp_operation_info_should_format_snmp_operation_data_correctly(
    function_name, locals_dict, expected_result
):
    """Test _format_snmp_operation_info function formats various operations correctly."""
    from zino.debug import _format_snmp_operation_info

    result = _format_snmp_operation_info(function_name, locals_dict)
    assert result == expected_result



"""Basic data models for keeping/serializing/deserializing Zino state"""

import datetime
import fnmatch
import logging
import pathlib
import re
from collections.abc import Generator
from enum import Enum
from ipaddress import IPv4Address, IPv6Address
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from pydantic import BaseModel, ConfigDict, Field

from zino.compat import StrEnum
from zino.time import now

if TYPE_CHECKING:
    from zino.state import ZinoState

IPAddress = Union[IPv4Address, IPv6Address]
AlarmType = Literal["yellow", "red"]
SubIndex = Union[None, int, IPAddress, AlarmType]
EventType = TypeVar("EventType", bound="Event")
CISCO_ENTERPRISE_ID = 9
JUNIPER_ENTERPRISE_ID = 2636

_logger = logging.getLogger(__name__)


class ClosedEventError(Exception):
    """Closed events cannot be reopened"""

    pass


class BGPStyle(StrEnum):
    """The set of allowable BGP styles"""

    JUNIPER = "juniper"
    CISCO = "cisco"
    GENERAL = "general"


class BGPAdminStatus(StrEnum):
    """The set of allowable BGP admin status"""

    RUNNING = "running"
    HALTED = "halted"
    START = "start"
    STOP = "stop"


class BGPOperState(StrEnum):
    """The set of allowable BGP oper states"""

    IDLE = "idle"
    CONNECT = "connect"
    ACTIVE = "active"
    OPENSENT = "opensent"
    OPENCONFIRM = "openconfirm"
    ESTABLISHED = "established"
    DOWN = "down"


class InterfaceState(StrEnum):
    """Enumerates allowable interface states.  Most of these values come from ifOperState from RFC 2863 (IF-MIB), but
    also adds internal Zino states that don't exist in the MIB.
    """

    ADMIN_DOWN = "adminDown"
    UP = "up"
    DOWN = "down"
    TESTING = "testing"
    UNKNOWN = "unknown"
    DORMANT = "dormant"
    NOT_PRESENT = "notPresent"
    LOWER_LAYER_DOWN = "lowerLayerDown"
    FLAPPING = "flapping"  # Zino-specific


class BFDSessState(StrEnum):
    """The set of allowable BFD session states"""

    ADMIN_DOWN = "adminDown"
    DOWN = "down"
    INIT = "init"
    UP = "up"
    FAILING = "failing"  # Cisco proprietary


class BFDState(BaseModel):
    """Keeps BFD state for an interface"""

    session_state: BFDSessState
    session_index: Optional[int] = None
    session_discr: Optional[int] = None
    session_addr: Optional[IPAddress] = None


class Port(BaseModel):
    """Keeps port state"""

    ifindex: int
    ifdescr: Optional[str] = None
    ifalias: Optional[str] = None
    state: Optional[InterfaceState] = None
    bfd_state: Optional[BFDState] = None


class BGPPeerSession(BaseModel):
    """Keeps a BGP peer session"""

    uptime: Optional[int] = 0
    admin_status: Optional[BGPAdminStatus] = None
    oper_state: Optional[BGPOperState] = None


class DeviceState(BaseModel):
    """Keep device state"""

    name: str
    addresses: set[IPAddress] = set()
    enterprise_id: Optional[int] = None
    ports: Dict[int, Port] = {}
    alarms: Optional[Dict[AlarmType, int]] = None
    boot_time: Optional[datetime.datetime] = None
    bgp_style: Optional[BGPStyle] = None
    bgp_peers: dict[IPAddress, BGPPeerSession] = dict()
    reachable_in_last_run: Optional[bool] = None

    # This is the remaining set of potential device attributes stored in device state by the original Zino code:
    # EventId
    # JNXalarms
    # RunsOn
    # bfdSessAddr
    # bfdSessAddrType
    # bfdSessDiscr
    # bfdSessState
    # firstFlap
    # flapping
    # flaps
    # isCisco
    # isJuniper
    # lastAge
    # lastFlap
    # portState
    # portToIfDescr
    # portToLocIfDescr
    # sawPeer

    @property
    def is_cisco(self):
        return self.enterprise_id == CISCO_ENTERPRISE_ID

    @property
    def is_juniper(self):
        return self.enterprise_id == JUNIPER_ENTERPRISE_ID

    def set_boot_time_from_uptime(self, uptime: int):
        """Calculates and sets the device boot time from a current uptime value.

        :param uptime: An uptime value in 100ths of a second
        """
        self.boot_time = now() - datetime.timedelta(seconds=uptime / 100)


class DeviceStates(BaseModel):
    """Keeps track of the state of all devices we have polled from"""

    devices: Dict[str, DeviceState] = {}

    def __getitem__(self, item) -> DeviceState:
        return self.devices[item]

    def __contains__(self, item):
        return item in self.devices

    def __len__(self):
        return len(self.devices)

    def get(self, device_name: str) -> DeviceState:
        """Returns a DeviceState object for device_name, creating a blank state object if none exists"""
        if device_name not in self:
            self.devices[device_name] = DeviceState(name=device_name)
        return self[device_name]


class LogEntry(BaseModel):
    """Event log entry attributes. These apply both for 'log' and 'history' lists"""

    timestamp: datetime.datetime = Field(default_factory=now)
    message: str

    def model_dump_legacy(self) -> List[str]:
        """Returns the contents of this log entry as a sequence of text lines.

        This sequence of text lines are formatted as expected in a multi-line response in the Zino legacy server
        protocol:

        The first line starts with a UNIX integer timestamp, followed by a space and then the first line of the entry
        text.  Each subsequent line of the entry text is prefixed by a space, which in the protocol signifies a
        continuation line.

        Example:
        >>> LogEntry(message="This is a\\nmulti-line entry").model_dump_legacy()
        ['1701171730 This is a', ' multiline entry']
        >>>
        """
        unix_timestamp = int(self.timestamp.timestamp())
        lines = [f" {line}" for line in self.message.splitlines()]
        lines[0] = f"{unix_timestamp}{lines[0]}"
        return lines


class EventState(Enum):
    """The set of allowable event states"""

    EMBRYONIC = "embryonic"
    OPEN = "open"
    WORKING = "working"
    WAITING = "waiting"
    CONFIRM = "confirm-wait"
    IGNORED = "ignored"
    CLOSED = "closed"


class ReachabilityState(Enum):
    """The set of allowed reachability states"""

    REACHABLE = "reachable"
    NORESPONSE = "no-response"


class Event(BaseModel):
    """Keeps track of event state"""

    # Allow populating fields by name or alias
    model_config = ConfigDict(populate_by_name=True)

    id: Optional[int] = None

    router: str
    type: Literal["Event"] = "Event"
    state: EventState = EventState.EMBRYONIC
    opened: datetime.datetime = Field(default_factory=now)
    updated: Optional[datetime.datetime] = None
    closed: Optional[datetime.datetime] = None
    priority: int = 100
    lastevent: Optional[str] = None

    log: List[LogEntry] = []
    history: List[LogEntry] = []

    # More-or-less optional event attrs (as guesstimated from the original Zino code)
    lasttrans: Optional[datetime.datetime] = None
    flaps: Optional[int] = None
    ac_down: Optional[datetime.timedelta] = None

    polladdr: Optional[IPAddress] = None

    @property
    def subindex(self) -> SubIndex:
        """Returns an identifier of a router subcomponent that this event refers to.

        This value is used to index an event uniquely, and the property itself is usually only called if the event index
        class needs to re-index all existing events - something that usually only happens when loading serialized
        state from disk.
        """
        return None

    def set_state(self, new_state: EventState, user: str = "monitor"):
        """Sets a new event state value, logging the change to the event history with a username"""
        if new_state == self.state:
            return
        if self.state == EventState.CLOSED:
            raise ClosedEventError

        old_state, self.state = self.state, new_state
        if (old_state, new_state) == (EventState.EMBRYONIC, EventState.OPEN):
            self.opened = now()
        self.add_history(f"state change {old_state.value} -> {new_state.value} ({user})")
        _logger.debug("id %s state %s -> %s by %s", self.id, old_state.value, new_state.value, user)

    def add_log(self, message: str) -> LogEntry:
        entry = LogEntry(message=message)
        self.log.append(entry)
        self.updated = entry.timestamp
        return entry

    def add_history(self, message: str) -> LogEntry:
        entry = LogEntry(message=message)
        self.history.append(entry)
        return entry

    def model_dump_simple_attrs(self) -> dict[str, str]:
        """Serializes the "simple" attributes of this event to the text format used by the legacy Zino protocol.

        "Simple" attributes are typically "single line" public attributes, i.e. anything but `log` or `history`.

        It would be nice to be able to attach custom serializers to fields, but we need to support multiple
        serialization formats (one for JSON state dumps, one for the legacy Zino protocol), and it's not clear how
        Pydantic can support that.
        """
        attrs = self.model_dump(mode="python", exclude={"log", "history"}, exclude_none=True, by_alias=True)
        return {attr.replace("_", "-"): self.zinoify_value(value) for attr, value in attrs.items()}

    @staticmethod
    def zinoify_value(value: Any) -> str:
        """Serializes an Event value into a plain string in the format expected by the legacy Zino protocol"""
        if isinstance(value, Enum):
            return str(value.value)
        if isinstance(value, datetime.datetime):
            return str(int(value.timestamp()))
        if isinstance(value, datetime.timedelta):
            return str(int(value.total_seconds()))
        return str(value)

    def get_changed_fields(self, other: "Event") -> List[str]:
        """Compares this Event to another Event and returns a list of names of attributes that are different"""
        return [field for field in self.model_fields if getattr(other, field) != getattr(self, field)]

    def dump_event_to_file(self, dir_name: str):
        """Dumps the full event to a file in JSON format"""
        pathlib.Path(dir_name).mkdir(parents=True, exist_ok=True)
        filename = f"{dir_name}/{self.id}.json"
        _logger.debug("dumping state to %s", filename)
        with open(filename, "w") as statefile:
            statefile.write(self.model_dump_json(exclude_none=True, indent=2))

    def is_down(self) -> bool:
        """Returns true if the event is in a down state.
        What is considered a "down state" depends on the type of event,
        but generally it means something like a device or a port is down.
        """
        return False


class FlapState(StrEnum):
    FLAPPING = "flapping"
    STABLE = "stable"


class PortStateEvent(Event):
    type: Literal["portstate"] = "portstate"
    port: Optional[str] = ""
    ifindex: Optional[int] = None
    portstate: Optional[InterfaceState] = None
    descr: Optional[str] = None
    flapstate: Optional[FlapState] = None
    reason: Optional[str] = None

    @property
    def subindex(self) -> SubIndex:
        return self.ifindex

    def is_down(self) -> bool:
        """Returns true if the event is in a down state.
        A PortStateEvent is considered "down" if the port it is related to is
        either down or in a flapping state.
        """
        return self.portstate in [InterfaceState.DOWN, InterfaceState.FLAPPING]


class BGPEvent(Event):
    type: Literal["bgp"] = "bgp"
    remote_address: Optional[IPAddress] = Field(default=None, alias="remote-addr")
    remote_as: Optional[int] = Field(default=None, alias="remote-AS")
    peer_uptime: Optional[int] = Field(default=None, alias="peer-uptime")
    operational_state: Optional[BGPOperState] = Field(default=None, alias="bgpOS")
    admin_status: Optional[BGPAdminStatus] = Field(default=None, alias="bgpAS")

    @property
    def subindex(self) -> SubIndex:
        return self.remote_address

    def is_down(self) -> bool:
        """Returns true if the event is in a down state.
        A BGPEvent is considered "down" if the operational state
        for the BGP session is not established.
        """
        return self.operational_state != BGPOperState.ESTABLISHED


class BFDEvent(Event):
    type: Literal["bfd"] = "bfd"
    ifindex: Optional[int] = None
    bfdstate: Optional[BFDSessState] = Field(default=None, alias="bfdState")
    bfdix: Optional[int] = Field(default=None, alias="bfdIx")
    bfddiscr: Optional[int] = Field(default=None, alias="bfdDiscr")
    bfdaddr: Optional[IPAddress] = Field(default=None, alias="bfdAddr")
    neigh_rdns: Optional[str] = Field(default=None, alias="Neigh-rDNS")

    @property
    def subindex(self) -> SubIndex:
        return self.ifindex


class ReachabilityEvent(Event):
    type: Literal["reachability"] = "reachability"
    reachability: Optional[ReachabilityState] = None

    def is_down(self) -> bool:
        """Returns true if the event is in a down state.
        A ReachabilityEvent is considered "down" if the device it is related to
        is not reachable.
        """
        return self.reachability == ReachabilityState.NORESPONSE


class AlarmEvent(Event):
    type: Literal["alarm"] = "alarm"
    alarm_type: Optional[AlarmType] = Field(default=None, alias="alarm-type")
    alarm_count: Optional[int] = Field(default=None, alias="alarm-count")

    @property
    def subindex(self) -> SubIndex:
        return self.alarm_type


class MatchType(StrEnum):
    """The set of allowable match types for PlannedMaintenanc objects"""

    REGEXP = "regexp"
    STR = "str"
    EXACT = "exact"
    INTF_REGEXP = "intf-regexp"


class PmType(StrEnum):
    """The set of allowable types for PlannedMaintenance objects"""

    PORTSTATE = "portstate"
    DEVICE = "device"


class PlannedMaintenance(BaseModel):
    # Allow populating fields by name or alias
    model_config = ConfigDict(populate_by_name=True)

    id: Optional[int] = None
    start_time: datetime.datetime = Field(alias="starttime")
    end_time: datetime.datetime = Field(alias="endtime")
    type: PmType
    match_type: MatchType
    match_device: Optional[str] = Field(default=None, alias="match_dev")
    match_expression: str = Field(alias="match_expr")
    log: List[LogEntry] = []
    event_ids: List[int] = []

    def start(self, state: "ZinoState"):
        events = self._get_or_create_events(state)
        for event in events:
            # get special handling of embryonic -> open transition first
            state.events.commit(event)
            event.state = EventState.IGNORED
            state.events.commit(event)
            self.event_ids.append(event.id)

    def end(self, state: "ZinoState"):
        for event_id in self.event_ids:
            event = state.events.checkout(event_id)
            event.state = EventState.OPEN
            state.events.commit(event)

    def add_log(self, message: str) -> LogEntry:
        entry = LogEntry(message=message)
        self.log.append(entry)
        return entry

    def details(self) -> str:
        """Returns a string with the details of the object.
        Format from zino1: $id $from_t $to_t $type $match_type [$match_dev] $match_expr
        """
        details = [
            str(int(attr.timestamp())) if isinstance(attr, datetime.datetime) else str(attr)
            for attr in [
                self.id,
                self.start_time,
                self.end_time,
                self.type,
                self.match_type,
                self.match_device,
                self.match_expression,
            ]
            if attr
        ]
        return " ".join(details)

    def matches_event(self, event: Event, state: "ZinoState") -> bool:
        """Returns true if `event` will be affected by this planned maintenance"""
        raise NotImplementedError

    def _get_or_create_events(self, state: "ZinoState") -> list[Event]:
        """Creates/gets events that are affected by the given starting planned
        maintenance
        """
        raise NotImplementedError

    def get_matching(self, state: "ZinoState") -> Iterator[Sequence[Union[str, int]]]:
        """Returns a list of matching devices or ports from Zino state.

        The number of elements of each sequence of the return value depends on the type of planned maintenance
        objects, but each entry should be suitable to join on space and output to the legacy API.
        """
        raise NotImplementedError


class DeviceMaintenance(PlannedMaintenance):
    type: Literal[PmType.DEVICE] = PmType.DEVICE

    def matches_event(self, event: Event, state: "ZinoState") -> bool:
        """Returns true if `event` will be affected by this planned maintenance"""
        if not isinstance(event, (ReachabilityEvent, AlarmEvent)):
            return False
        device = state.devices[event.router]
        return self.matches_device(device)

    def matches_device(self, device: DeviceState) -> bool:
        """Returns true if ReachabilityEvents and AlarmEvents related to `device`
        would be affected by this planned maintenance
        """
        if self.match_type == MatchType.REGEXP:
            return regex_search(self.match_expression, device.name)
        if self.match_type == MatchType.STR:
            return string_match(self.match_expression, device.name)
        if self.match_type == MatchType.EXACT:
            return self.match_expression == device.name
        return False

    def get_matching(self, state: "ZinoState") -> Iterator[Sequence[Union[str, int]]]:
        """Returns a list of matching devices from Zino state.

        Each element is a sequence of (pm_id, "device", device_name)
        """
        for device in state.devices.devices.values():
            if self.matches_device(device):
                yield self.id, self.type, device.name

    def _get_or_create_events(self, state: "ZinoState") -> list[Event]:
        """Creates/gets events that are affected by the given starting planned
        maintenance
        """
        events = []
        # all devices that the pm should affect
        devices = (device for device in state.devices.devices.values() if self.matches_device(device))
        for device in devices:
            reachability_event = state.events.get_or_create_event(device.name, None, ReachabilityEvent)
            yellow_event = state.events.get_or_create_event(device.name, "yellow", AlarmEvent)
            yellow_event.alarm_type = "yellow"
            red_event = state.events.get_or_create_event(device.name, "red", AlarmEvent)
            red_event.alarm_type = "red"
            for event in (reachability_event, yellow_event, red_event):
                events.append(event)
        return events


class PortStateMaintenance(PlannedMaintenance):
    type: Literal[PmType.PORTSTATE] = PmType.PORTSTATE

    def matches_event(self, event: Event, state: "ZinoState") -> bool:
        """Returns true if `event` will be affected by this planned maintenance"""
        if not isinstance(event, PortStateEvent):
            return False
        device = state.devices[event.router]
        port = device.ports[event.ifindex]
        return self.matches_portstate(device, port)

    def matches_portstate(self, device: DeviceState, port: Port) -> bool:
        """Returns true if PortstateEvents related to `port` on `device`
        would be affected by this planned maintenance
        """
        if self.match_type == MatchType.REGEXP:
            return regex_search(self.match_expression, port.ifalias)
        if self.match_type == MatchType.STR:
            return string_match(self.match_expression, port.ifalias)
        if self.match_type == MatchType.INTF_REGEXP:
            if regex_search(self.match_device, device.name):
                return regex_search(self.match_expression, port.ifdescr)
        return False

    def get_matching(self, state: "ZinoState") -> Iterator[Sequence[Union[str, int]]]:
        """Returns a list of matching devices from Zino state.

        Each element is a sequence of (pm_id, "portstate", device_name, ifIndex, ifDescr, f"({ifAlias})")
        """
        for device in state.devices.devices.values():
            for port in device.ports.values():
                if self.matches_portstate(device, port):
                    yield self.id, self.type, device.name, port.ifindex, port.ifdescr, f"({port.ifalias})"

    def _get_or_create_events(self, state: "ZinoState") -> list[Event]:
        events = []
        for device, port in self._get_matching_ports(state):
            event = state.events.get_or_create_event(device.name, port.ifindex, PortStateEvent)
            event.ifindex = port.ifindex
            events.append(event)
        return events

    def _get_matching_ports(self, state: "ZinoState") -> Generator[tuple[DeviceState, Port], None, None]:
        for device in state.devices.devices.values():
            for port in device.ports.values():
                if self.matches_portstate(device, port):
                    yield (device, port)


def string_match(pattern: str, string: str) -> bool:
    """This should behave like tcl string match https://wiki.tcl-lang.org/page/string+match
    Returns true if `string` matches `pattern`.
    """
    return fnmatch.fnmatch(string, pattern)


def regex_search(pattern: str, string: str) -> bool:
    """Matches `string` against regex expression `pattern`.
    Returns true if there is a match.
    """
    return bool(re.search(pattern, string))



import logging
import re
from ipaddress import ip_address
from typing import Any, List, Optional

from zino.statemodels import IPAddress
from zino.tasks.task import Task

_logger = logging.getLogger(__name__)


IGNORE_ADDRS = {
    ip_address(ip)
    for ip in (
        # UNINETTs anycast PIM RP addresses (should be configurable)
        "128.39.0.85",
        "128.39.0.2",
        # SUNET PIM RP address
        "193.10.80.229",
        # NORDUnet PIM RP addresses
        "193.10.251.1",
        "109.105.96.128",
    )
}

IGNORE_ADDR_PATTERNS = [
    re.compile(pat)
    for pat in (
        r"^10\.",
        r"^127\.",
        r"^128\.0\.",
        r"^172\.(1[6-9]|2[0-9]|3[0-1])\.",
        r"^192\.168\.",
    )
]


class AddressMapTask(Task):
    """Fetches and stores mappings between router names and router interface addresses in state data.

    These mappings can be used for things like identifying the originating router from SNMP traps that are sent using
    a different source address than the management interface Zino normally talks to.
    """

    async def run(self):
        addresses = await self._get_addrs()
        _logger.debug("found addresses for %s: %r", self.device.name, addresses)
        self._update_address_maps(addresses)

    async def _get_addrs(self) -> set[IPAddress]:
        result = await self.snmp.bulkwalk("IP-MIB", "ipAdEntAddr")
        addresses = (validate_ipaddr(r.value) for r in result)
        return set(addr for addr in addresses if addr)

    def _update_address_maps(self, addresses: List[IPAddress]):
        state = self.state
        for address in addresses:
            if self.is_ignored(address):
                continue
            if address not in state.addresses:
                _logger.info("%s adds address %s", self.device.name, address)
            elif state.addresses[address] != self.device.name:
                _logger.debug("Home of %s changed from %s to %s", address, state.addresses[address], self.device.name)

            state.addresses[address] = self.device.name

        self.device_state.addresses = addresses

        missing_addresses = set(
            address
            for address, router in state.addresses.items()
            if router == self.device.name and address not in addresses
        )
        if missing_addresses:
            _logger.info("%s no longer has these addresses: %r", self.device.name, missing_addresses)
            for address in missing_addresses:
                del state.addresses[address]

    @staticmethod
    def is_ignored(address: IPAddress) -> bool:
        """Returns True if a particular address should be ignored.

        For the time being, the address lists and patterns to ignore are hard-coded in this module, as they were in
        Zino 1, but they should later be moved to a configuration file.
        """
        return address in IGNORE_ADDRS or any(pattern.match(str(address)) for pattern in IGNORE_ADDR_PATTERNS)


def validate_ipaddr(address: Any) -> Optional[IPAddress]:
    """Converts and returns address as a IPv4Address or IPv6Address object.  Returns None if the address is invalid."""
    try:
        return ip_address(address)
    except ValueError:
        return None



#
# PySNMP MIB module OLD-CISCO-SYSTEM-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file:///home/mvold/mibs/cisco/OLD-CISCO-SYSTEM-MIB.my
# Produced by pysmi-0.3.4 at Thu Jul 18 11:03:56 2024
# On host agrajag platform Linux version 6.6.37 by user mvold
# Using Python version 3.11.9 (main, Apr  2 2024, 08:25:04) [GCC 13.2.0]
#
Integer, ObjectIdentifier, OctetString = mibBuilder.importSymbols("ASN1", "Integer", "ObjectIdentifier", "OctetString")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
SingleValueConstraint, ValueSizeConstraint, ValueRangeConstraint, ConstraintsUnion, ConstraintsIntersection = mibBuilder.importSymbols("ASN1-REFINEMENT", "SingleValueConstraint", "ValueSizeConstraint", "ValueRangeConstraint", "ConstraintsUnion", "ConstraintsIntersection")
local, = mibBuilder.importSymbols("CISCO-SMI", "local")
NotificationGroup, ModuleCompliance = mibBuilder.importSymbols("SNMPv2-CONF", "NotificationGroup", "ModuleCompliance")
Bits, Counter64, IpAddress, iso, Gauge32, Integer32, ModuleIdentity, Unsigned32, MibIdentifier, ObjectIdentity, MibScalar, MibTable, MibTableRow, MibTableColumn, NotificationType, Counter32, TimeTicks = mibBuilder.importSymbols("SNMPv2-SMI", "Bits", "Counter64", "IpAddress", "iso", "Gauge32", "Integer32", "ModuleIdentity", "Unsigned32", "MibIdentifier", "ObjectIdentity", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "NotificationType", "Counter32", "TimeTicks")
TextualConvention, DisplayString = mibBuilder.importSymbols("SNMPv2-TC", "TextualConvention", "DisplayString")
lsystem = MibIdentifier((1, 3, 6, 1, 4, 1, 9, 2, 1))
romId = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 1), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: romId.setStatus('mandatory')
whyReload = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 2), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: whyReload.setStatus('mandatory')
hostName = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 3), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: hostName.setStatus('mandatory')
domainName = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 4), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: domainName.setStatus('mandatory')
authAddr = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 5), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: authAddr.setStatus('mandatory')
bootHost = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 6), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bootHost.setStatus('mandatory')
netConfigAddr = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 48), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: netConfigAddr.setStatus('mandatory')
netConfigName = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 49), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: netConfigName.setStatus('mandatory')
netConfigSet = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 50), DisplayString()).setMaxAccess("writeonly")
if mibBuilder.loadTexts: netConfigSet.setStatus('mandatory')
hostConfigAddr = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 51), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: hostConfigAddr.setStatus('obsolete')
hostConfigName = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 52), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: hostConfigName.setStatus('obsolete')
hostConfigSet = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 53), DisplayString()).setMaxAccess("writeonly")
if mibBuilder.loadTexts: hostConfigSet.setStatus('obsolete')
writeMem = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 54), Integer32()).setMaxAccess("writeonly")
if mibBuilder.loadTexts: writeMem.setStatus('mandatory')
writeNet = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 55), DisplayString()).setMaxAccess("writeonly")
if mibBuilder.loadTexts: writeNet.setStatus('mandatory')
ciscoContactInfo = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 61), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: ciscoContactInfo.setStatus('mandatory')
netConfigProto = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 70), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: netConfigProto.setStatus('mandatory')
hostConfigProto = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 71), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: hostConfigProto.setStatus('mandatory')
sysConfigAddr = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 72), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysConfigAddr.setStatus('mandatory')
sysConfigName = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 73), DisplayString()).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysConfigName.setStatus('mandatory')
sysConfigProto = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 74), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3))).clone(namedValues=NamedValues(("tftp", 1), ("rom", 2), ("flash", 3)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: sysConfigProto.setStatus('mandatory')
sysClearARP = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 75), Integer32()).setMaxAccess("writeonly")
if mibBuilder.loadTexts: sysClearARP.setStatus('mandatory')
sysClearInt = MibScalar((1, 3, 6, 1, 4, 1, 9, 2, 1, 76), Integer32()).setMaxAccess("writeonly")
if mibBuilder.loadTexts: sysClearInt.setStatus('mandatory')
mibBuilder.exportSymbols("OLD-CISCO-SYSTEM-MIB", domainName=domainName, whyReload=whyReload, netConfigName=netConfigName, ciscoContactInfo=ciscoContactInfo, sysConfigAddr=sysConfigAddr, sysConfigProto=sysConfigProto, romId=romId, hostConfigSet=hostConfigSet, netConfigSet=netConfigSet, hostConfigName=hostConfigName, netConfigAddr=netConfigAddr, lsystem=lsystem, hostConfigProto=hostConfigProto, writeMem=writeMem, sysClearARP=sysClearARP, bootHost=bootHost, netConfigProto=netConfigProto, hostConfigAddr=hostConfigAddr, authAddr=authAddr, sysConfigName=sysConfigName, sysClearInt=sysClearInt, writeNet=writeNet, hostName=hostName)



"""SNMP trap 'daemon' for Zino 2 implemented using PySNMP"""

import asyncio
import logging
from ipaddress import ip_address
from typing import Dict, Optional

from pyasn1.type.base import SimpleAsn1Type
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity import config
from pysnmp.entity.rfc3413 import ntfrcv
from pysnmp.proto.rfc1902 import ObjectName
from pysnmp.smi import view
from pysnmp.smi.rfc1902 import ObjectIdentity, ObjectType

import zino.state
from zino.config.models import PollDevice
from zino.oid import OID
from zino.snmp.pysnmp_backend import get_new_snmp_engine, mib_value_to_python
from zino.trapd.base import (
    TrapMessage,
    TrapOriginator,
    TrapReceiverBase,
    TrapVarBind,
)

_logger = logging.getLogger(__name__)


class TrapReceiver(TrapReceiverBase):
    """Zino Adapter for SNMP trap reception using PySNMP.

    A major difference to Zino 1 is that this receiver must explicitly be configured with SNMP community strings that
    will be accepted.  Zino 1 accepts traps with any community string, as long as their origin is any one of the
    devices configured in the pollfile.  However, PySNMP places heavy emphasis on being standards compliant,
    and will not even pass on traps to our callbacks unless they match the authorization config for the SNMP engine.
    """

    def __init__(
        self,
        address: str = "0.0.0.0",
        port: int = 162,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        state: Optional[zino.state.ZinoState] = None,
        polldevs: Optional[Dict[str, PollDevice]] = None,
    ):
        super().__init__(address, port, loop, state, polldevs)
        self.snmp_engine = get_new_snmp_engine()

    async def open(self):
        """Opens the UDP transport socket and starts receiving traps"""
        self.transport = udp.UdpTransport(loop=self.loop).openServerMode((self.address, self.port))
        # This attribute needs to be awaited to ensure the socket is really opened,
        # unsure of why PySNMP doesn't do this, or how else it's supposed to be achieved
        await self.transport._lport
        _logger.info("Listening for incoming SNMP traps on %r", (self.address, self.port))
        config.addTransport(self.snmp_engine, udp.domainName + (1,), self.transport)

        ntfrcv.NotificationReceiver(self.snmp_engine, self.trap_received)
        self.snmp_engine.transportDispatcher.jobStarted(1)  # this job would never finish

    def close(self):
        """Closes the running SNMP engine and its associated ports"""
        self.snmp_engine.transportDispatcher.closeDispatcher()
        self.transport.closeTransport()

    def add_community(self, community: str):
        """Adds a new community string that will be accepted on incoming packets"""
        super().add_community(community)
        config.addV1System(self.snmp_engine, str(len(self._communities)), community)

    def trap_received(self, snmp_engine, state_reference, context_engine_id, context_name, var_binds, callback_context):
        """Callback function that receives all matched trap messages on PySNMP's incoming transport socket"""
        transport_domain, transport_address = snmp_engine.msgAndPduDsp.getTransportInfo(state_reference)
        sender_address, sender_port = transport_address
        sender_address = ip_address(sender_address)

        router = self._lookup_device(sender_address)
        if not router:
            _logger.info("ignored trap from %s (not a box we monitor?)", sender_address)
            return

        _logger.debug(
            'Trap from %s (%s) ContextEngineId "%s", ContextName "%s"',
            router.name,
            sender_address,
            context_engine_id.prettyPrint(),
            context_name.prettyPrint(),
        )
        origin = TrapOriginator(address=sender_address, port=sender_port, device=router)
        trap = TrapMessage(agent=origin)
        snmp_trap_oid: ObjectName = None
        for var, raw_value in var_binds:
            mib, label, instance, raw_value = self._resolve_varbind(var, raw_value)
            _logger.debug("(%r, %r, %s) = %s", mib, label, instance, raw_value.prettyPrint())
            try:
                value = mib_value_to_python(raw_value)
            except Exception:  # noqa
                value = None
            trap.variables.append(TrapVarBind(OID(var), mib, label, instance, raw_value, value))
            if label == "snmpTrapOID":
                snmp_trap_oid = raw_value

        if not self._verify_trap(trap):
            return

        # TODO do some time calculations, but ask Håvard what the deal is with RestartTime vs. BootTime

        trap.mib, trap.name, _ = self._resolve_object_name(snmp_trap_oid)
        asyncio.ensure_future(self.dispatch_trap(trap))

    @staticmethod
    def _verify_trap(trap: TrapMessage) -> bool:
        device = trap.agent.device.name if trap.agent.device else "N/A"
        if "snmpTrapOID" not in trap:
            _logger.error("Trap from %s did not contain a snmpTrapOID value, ignoring", device)
            return False

        if "sysUpTime" not in trap:
            _logger.error("Trap from %s did not contain a sysUpTime value, ignoring", device)
            return False

        return True

    def _resolve_object_name(self, object_name: ObjectName) -> tuple[str, str, OID]:
        """Raises MibNotFoundError if oid in `object_name` can not be found"""
        engine = self.snmp_engine
        controller = engine.getUserContext("mibViewController")
        if not controller:
            controller = view.MibViewController(engine.getMibBuilder())
        mib, label, instance = controller.getNodeLocation(object_name)
        return mib, label, OID(instance) if instance else None

    def _resolve_varbind(self, name: ObjectName, value: SimpleAsn1Type) -> tuple[str, str, OID, SimpleAsn1Type]:
        """Resolves a varbind name and value to a MIB, label, instance, and value.  The value will be interpreted
        according to the resolved MIB object.

        Raises MibNotFoundError if the object's MIB cannot be resolved.
        """
        engine = self.snmp_engine
        controller = engine.getUserContext("mibViewController")
        if not controller:
            controller = view.MibViewController(engine.getMibBuilder())

        name, value = ObjectType(ObjectIdentity(name), value).resolveWithMib(controller)
        mib, label, instance = name.getMibSymbol()
        return mib, label, instance, value



import logging
from dataclasses import dataclass, replace
from ipaddress import ip_address
from typing import Iterable, Optional

from zino.snmp.base import SparseWalkResponse
from zino.statemodels import (
    BGPAdminStatus,
    BGPEvent,
    BGPOperState,
    BGPPeerSession,
    BGPStyle,
    IPAddress,
)
from zino.tasks.task import Task

_logger = logging.getLogger(__name__)

TIME_BEFORE_OPER_DOWN_ALERT = 600

JUNIPER_TRANSLATION_MAP = [
    ("peer_state", "jnxBgpM2PeerState"),
    ("peer_admin_status", "jnxBgpM2PeerStatus"),
    ("peer_remote_address", "jnxBgpM2PeerRemoteAddr"),
    ("peer_remote_as", "jnxBgpM2PeerRemoteAs"),
    ("peer_fsm_established_time", "jnxBgpM2PeerFsmEstablishedTime"),
]
CISCO_TRANSLATION_MAP = [
    ("peer_state", "cbgpPeer2State"),
    ("peer_admin_status", "cbgpPeer2AdminStatus"),
    ("peer_remote_address", "cbgpPeer2RemoteAddr"),
    ("peer_remote_as", "cbgpPeer2RemoteAs"),
    ("peer_fsm_established_time", "cbgpPeer2FsmEstablishedTime"),
]
GENERAL_TRANSLATION_MAP = [
    ("peer_state", "bgpPeerState"),
    ("peer_admin_status", "bgpPeerAdminStatus"),
    ("peer_remote_address", "bgpPeerRemoteAddr"),
    ("peer_remote_as", "bgpPeerRemoteAs"),
    ("peer_fsm_established_time", "bgpPeerFsmEstablishedTime"),
]
LOCAL_AS_OBJECTS = {
    "juniper": ("BGP4-V2-MIB-JUNIPER", "jnxBgpM2PeerLocalAs"),
    "cisco": ("CISCO-BGP4-MIB", "cbgpLocalAs"),
    "general": ("BGP4-MIB", "bgpLocalAs"),
}
BUGGY_REMOTE_ADDRESSES = [
    # Bug in JunOS -- info from IPv6 BGP sessions spill over
    ip_address("0.0.0.0"),
    # Bug in earlier Cisco IOS, info from elsewhere (IPv6?) spills over
    ip_address("32.1.7.0"),
]


@dataclass
class BaseBGPRow:
    peer_state: BGPOperState
    peer_admin_status: BGPAdminStatus
    peer_remote_address: IPAddress
    peer_remote_as: int
    peer_fsm_established_time: int

    def __post_init__(self):
        """Coerces incoming values to expected types"""
        self.peer_state = BGPOperState(self.peer_state)
        self.peer_admin_status = BGPAdminStatus(self.peer_admin_status)
        self.peer_remote_address = ip_address(self.peer_remote_address)
        self.peer_remote_as = int(self.peer_remote_as)
        self.peer_fsm_established_time = int(self.peer_fsm_established_time)


class BGPStateMonitorTask(Task):
    """Fetches and stores state information about external BGP sessions."""

    async def run(self):
        if not self.device.do_bgp:
            _logger.debug("Skipping BGP scanning for %s due to config", self.device.name)
            return

        bgp_style = await self._get_bgp_style()
        if bgp_style != self.device_state.bgp_style:
            _logger.debug(
                f"Router {self.device_state.name} changed its BGP style from '{self.device_state.bgp_style}' to "
                f"'{bgp_style}'"
            )
            self.device_state.bgp_style = bgp_style
        if not bgp_style:
            return

        local_as = await self._get_local_as(bgp_style=bgp_style)
        uptime = await self._get_uptime()
        if not local_as or not uptime:
            return

        if bgp_style == BGPStyle.JUNIPER:
            bgp_info = await self._get_juniper_bgp_info()
        elif bgp_style == BGPStyle.CISCO:
            bgp_info = await self._get_cisco_bgp_info()
        elif bgp_style == BGPStyle.GENERAL:
            bgp_info = await self._get_general_bgp_info()

        if not bgp_info:
            return

        for result in bgp_info:
            self._update_single_bgp_entry(data=result, local_as=local_as, uptime=uptime)

    async def _get_bgp_style(self) -> Optional[BGPStyle]:
        if await self.snmp.subtree_is_supported("BGP4-V2-MIB-JUNIPER", "jnxBgpM2"):
            return BGPStyle.JUNIPER

        if await self.snmp.subtree_is_supported("CISCO-BGP4-MIB", "cbgpPeer2Table"):
            return BGPStyle.CISCO

        if await self.snmp.subtree_is_supported("BGP4-MIB", "bgp"):
            return BGPStyle.GENERAL

        return None

    async def _get_local_as(self, bgp_style: BGPStyle) -> Optional[int]:
        try:
            mib, object_name = LOCAL_AS_OBJECTS[bgp_style]
        except KeyError:
            return
        response = await self.snmp.getnext(mib, object_name)
        if self.snmp.is_in_scope(response, (mib, object_name)):
            return response.value
        else:
            _logger.info(f"router {self.device.name} misses {object_name}")

    async def _get_juniper_bgp_info(self) -> Optional[list[BaseBGPRow]]:
        variables = (
            "jnxBgpM2PeerState",
            "jnxBgpM2PeerStatus",
            "jnxBgpM2PeerRemoteAddr",
            "jnxBgpM2PeerRemoteAs",
            "jnxBgpM2PeerFsmEstablishedTime",
        )

        juniper_bgp_info = await self._get_bgp_info(mib_name="BGP4-V2-MIB-JUNIPER", variables=variables)

        if not juniper_bgp_info:
            return None

        juniper_bgp_info = self._transform_variables_from_specific_to_general(
            bgp_info=juniper_bgp_info, bgp_style=BGPStyle.JUNIPER
        )

        return juniper_bgp_info

    async def _get_cisco_bgp_info(self) -> Optional[list[BaseBGPRow]]:
        variables = (
            "cbgpPeer2State",
            "cbgpPeer2AdminStatus",
            "cbgpPeer2RemoteAs",
            "cbgpPeer2FsmEstablishedTime",
        )

        cisco_bgp_info = await self._get_bgp_info(mib_name="CISCO-BGP4-MIB", variables=variables)

        if not cisco_bgp_info:
            return None

        for oid, result in cisco_bgp_info.items():
            _addr_type, _addr_len, addr = oid[0], oid[1], oid[2:]
            result["cbgpPeer2RemoteAddr"] = ip_address(bytes(addr))

        cisco_bgp_info = self._transform_variables_from_specific_to_general(
            bgp_info=cisco_bgp_info, bgp_style=BGPStyle.CISCO
        )

        return cisco_bgp_info

    async def _get_general_bgp_info(self) -> Optional[list[BaseBGPRow]]:
        variables = (
            "bgpPeerState",
            "bgpPeerAdminStatus",
            "bgpPeerRemoteAddr",
            "bgpPeerRemoteAs",
            "bgpPeerFsmEstablishedTime",
        )

        general_bgp_info = await self._get_bgp_info(mib_name="BGP4-MIB", variables=variables)

        if not general_bgp_info:
            return None

        general_bgp_info = self._transform_variables_from_specific_to_general(
            bgp_info=general_bgp_info, bgp_style=BGPStyle.GENERAL
        )

        return general_bgp_info

    async def _get_bgp_info(self, mib_name: str, variables: Iterable[str]) -> Optional[SparseWalkResponse]:
        bgp_info = await self.snmp.sparsewalk(
            *((mib_name, var) for var in variables),
            max_repetitions=3,
        )
        return bgp_info

    def _transform_variables_from_specific_to_general(
        self, bgp_info: SparseWalkResponse, bgp_style: BGPStyle
    ) -> Optional[list[BaseBGPRow]]:
        if bgp_style == BGPStyle.JUNIPER:
            translation = JUNIPER_TRANSLATION_MAP
        elif bgp_style == BGPStyle.CISCO:
            translation = CISCO_TRANSLATION_MAP
        elif bgp_style == BGPStyle.GENERAL:
            translation = GENERAL_TRANSLATION_MAP

        generalized_bgp_info = {key: {} for key in bgp_info.keys()}

        for oid, result in bgp_info.items():
            missing_variables = []
            for general_name, specific_name in translation:
                try:
                    generalized_bgp_info[oid][general_name] = result[specific_name]
                except KeyError:
                    missing_variables.append(specific_name)

        if missing_variables:
            _logger.info(f"router {self.device.name} misses BGP variables ({missing_variables})")
            return None

        return [BaseBGPRow(**result) for result in generalized_bgp_info.values()]

    def _update_single_bgp_entry(self, data: BaseBGPRow, local_as: int, uptime: int):
        if data.peer_remote_address in BUGGY_REMOTE_ADDRESSES:
            return

        # Internal bgp sessions are not observed
        if local_as == data.peer_remote_as:
            return

        if data.peer_state == BGPOperState.ESTABLISHED:
            self._update_established_peer(data, uptime)
        else:
            self._update_nonestablished_peer(data, uptime)

        # Update device state with BGP session
        self.device_state.bgp_peers[data.peer_remote_address] = BGPPeerSession(
            uptime=data.peer_fsm_established_time, admin_status=data.peer_admin_status, oper_state=data.peer_state
        )

    def _update_established_peer(self, data: BaseBGPRow, uptime: int):
        saved_bgp_peer_session = self.device_state.bgp_peers.get(data.peer_remote_address)
        if saved_bgp_peer_session and uptime >= saved_bgp_peer_session.uptime > data.peer_fsm_established_time:
            self._bgp_external_reset(data)
            _logger.debug(f"Noted external reset for {self.device_state.name}: {data.peer_remote_address}")
        else:
            event = self.state.events.get(self.device.name, data.peer_remote_address, BGPEvent)
            if event and event.operational_state != BGPOperState.ESTABLISHED:
                self._bgp_external_reset(data)
                _logger.debug(f"BGP session up for {self.device_state.name}: {data.peer_remote_address}")

    def _update_nonestablished_peer(self, data: BaseBGPRow, uptime: int):
        saved_bgp_peer_session = self.device_state.bgp_peers.get(data.peer_remote_address)
        if data.peer_admin_status in [BGPAdminStatus.STOP, BGPAdminStatus.HALTED]:
            if not saved_bgp_peer_session or saved_bgp_peer_session.admin_status != data.peer_admin_status:
                self._bgp_admin_down(data)
                _logger.debug(
                    f"Router {self.device_state.name} peer {data.peer_remote_address} AS {data.peer_remote_as} "
                    f"admin-down"
                )
        else:
            self._update_peer_with_admin_status_start_or_running(data, saved_bgp_peer_session, uptime)

    def _update_peer_with_admin_status_start_or_running(
        self, data: BaseBGPRow, saved_bgp_peer_session: BGPPeerSession, uptime: int
    ):
        if not saved_bgp_peer_session or saved_bgp_peer_session.admin_status != data.peer_admin_status:
            self._bgp_admin_up(data)
        if not saved_bgp_peer_session or saved_bgp_peer_session.oper_state == BGPOperState.ESTABLISHED:
            # First verify that we've been up longer than the required time before we flag it as an alert
            if uptime > TIME_BEFORE_OPER_DOWN_ALERT:
                self._bgp_oper_down(data)
                _logger.debug(
                    f"Router {self.device_state.name} peer {data.peer_remote_address} AS {data.peer_remote_as} "
                    f"is {data.peer_state} (down)",
                )
            else:
                _logger.debug(
                    f"Router {self.device_state.name} peer {data.peer_remote_address} AS {data.peer_remote_as} "
                    f"is {data.peer_state} (down), but uptime = {uptime}",
                )

    def _bgp_external_reset(self, data: BaseBGPRow):
        event = self.state.events.get_or_create_event(self.device.name, data.peer_remote_address, BGPEvent)
        event = self._update_bgp_event(event=event, data=data, last_event="peer was reset (now up)")

        log = f"{event.router} peer {data.peer_remote_address} AS {data.peer_remote_as} was reset (now up)"
        _logger.info(log)
        event.add_log(log)

        self.state.events.commit(event=event)

    def _bgp_admin_down(self, data: BaseBGPRow):
        event = self.state.events.get_or_create_event(self.device.name, data.peer_remote_address, BGPEvent)

        if event.admin_status == data.peer_admin_status:
            return

        copied_data = replace(data, peer_state="down", peer_fsm_established_time=0)
        event = self._update_bgp_event(event=event, data=copied_data, last_event="peer is admin turned off")

        log = (
            f"{event.router} peer {data.peer_remote_address} AS {data.peer_remote_as} is admin turned off "
            f"({data.peer_admin_status})"
        )
        _logger.info(log)
        event.add_log(log)

        self.state.events.commit(event=event)

    def _bgp_admin_up(self, data: BaseBGPRow):
        event = self.state.events.get_or_create_event(self.device.name, data.peer_remote_address, BGPEvent)

        # No previous event, so no need to notify or event already up to date
        if event.id is None or event.admin_status == data.peer_admin_status:
            return

        copied_data = replace(data, peer_fsm_established_time=0)
        event = self._update_bgp_event(event=event, data=copied_data, last_event="peer is now admin turned on")

        log = (
            f"{event.router} peer {data.peer_remote_address} AS {data.peer_remote_as} is now admin turned on "
            f"({data.peer_admin_status})"
        )
        _logger.info(log)
        event.add_log(log)

        self.state.events.commit(event=event)

    def _bgp_oper_down(self, data: BaseBGPRow):
        event = self.state.events.get_or_create_event(self.device.name, data.peer_remote_address, BGPEvent)

        if event.operational_state == BGPOperState.DOWN:
            return

        copied_data = replace(data, peer_state=BGPOperState.DOWN)
        event = self._update_bgp_event(event=event, data=copied_data, last_event="peer is down")

        log = (
            f"{event.router} peer {data.peer_remote_address} AS {data.peer_remote_as} is down "
            f"({data.peer_admin_status})"
        )
        _logger.info(log)
        event.add_log(log)

        self.state.events.commit(event=event)

    def _update_bgp_event(self, event: BGPEvent, data: BaseBGPRow, last_event: str) -> BGPEvent:
        """Updates a given BGP event with the given BGP data"""

        event.operational_state = data.peer_state
        event.admin_status = data.peer_admin_status
        event.remote_address = data.peer_remote_address
        event.remote_as = data.peer_remote_as
        event.peer_uptime = data.peer_fsm_established_time
        event.polladdr = self.device.address
        event.priority = self.device.priority
        event.lastevent = last_event

        return event



import asyncio


async def send_trap_externally(*args: str, port: int = 1162):
    """Uses the snmptrap command line program to send a trap to the local trap receiver test instance on port 1162.

    :param args: The arguments to pass to the snmptrap command line program, see `man snmptrap` for details on the
                 rather esoteric syntax.
    """
    args = " ".join(args)
    proc = await asyncio.create_subprocess_shell(f"snmptrap -v 2c -c public localhost:{port} '' {args}")
    await proc.communicate()
    assert proc.returncode == 0, "snmptrap command exited with error"



import asyncio
import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, NamedTuple, Optional, Set

import zino.state
from zino.config.models import PollDevice
from zino.oid import OID
from zino.statemodels import DeviceState, IPAddress

TrapType = tuple[str, str]  # A mib name and a corresponding trap symbolic name

_logger = logging.getLogger(__name__)


class TrapVarBind(NamedTuple):
    """Describes a single trap varbind as high-level as possible, but with low level details available as well"""

    oid: OID
    mib: str
    var: str
    instance: OID
    raw_value: Any
    value: Any


class TrapOriginator(NamedTuple):
    """Describes the originating SNMP agent of a trap message in Zino terms"""

    address: IPAddress
    port: int
    device: Optional[DeviceState] = None


@dataclass
class TrapMessage:
    """Describes an incoming trap message in the simplest possible terms needed for Zino usage"""

    agent: TrapOriginator
    mib: Optional[str] = None
    name: Optional[str] = None
    variables: List[TrapVarBind] = field(default_factory=list)

    def __str__(self):
        variables = [f"{v.mib}::{v.var}{v.instance or ''}={v.value or v.raw_value}" for v in self.variables]
        variables = ", ".join(variables)
        trap_name = f"{self.mib}::{self.name}" if self.mib and self.name else "unknown"
        return f"<Trap from {self.agent.device.name} ({trap_name}): {variables}>"

    def __contains__(self, label) -> bool:
        for var in self.variables:
            if var.var == label:
                return True
        return False

    def get_all(self, label: str) -> List[TrapVarBind]:
        """Returns all contained variables with the given label"""
        return [var for var in self.variables if var.var == label]


class TrapObserver:
    """Defines a valid protocol for SNMP trap observers.

    A trap observer that directly subclasses this protocol can expect to be automatically registered by Zino as an
    observer for any trap it declares in its `WANTED_TRAPS` attribute.
    """

    WANTED_TRAPS: Set[TrapType] = set()

    def __init__(
        self,
        state: zino.state.ZinoState,
        polldevs: Optional[Dict[str, PollDevice]] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        """Initializes a new trap observer with a reference to Zino's state.

        Optionally also receives a reference to the current polldevs configuration and the running event loop.
        """
        self.state = state
        self.polldevs: Dict[str, PollDevice] = polldevs or zino.state.polldevs
        self.loop = loop if loop else asyncio.get_event_loop()

    async def handle_trap(self, trap: TrapMessage) -> Optional[bool]:
        """A trap observer receives a trap message and an optional event loop reference.  The event loop reference
        may be useful if the observer needs to run async actions as part of its trap processing.  If the trap
        observer returns a true-ish value, the trap dispatcher will offer the same trap to more subscribers.  If a
        false-ish value is returned, the trap dispatcher will stop processing this trap.
        """
        ...


class TrapReceiverBase:
    """Base class for Zino SNMP trap receiver back-ends.  Contains common
    functionality for all trap receivers, not specific to any underlying library.
    """

    def __init__(
        self,
        address: str = "0.0.0.0",
        port: int = 162,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        state: Optional[zino.state.ZinoState] = None,
        polldevs: Optional[Dict[str, PollDevice]] = None,
    ):
        self.address = address
        self.port = port
        self.loop = loop if loop else asyncio.get_running_loop()
        self.state = state or zino.state.ZinoState()
        self.polldevs = polldevs if polldevs is not None else {}
        self._communities = set()
        self._observers: dict[TrapType, List[TrapObserver]] = {}
        self._auto_subscribed_observers = set()

    async def open(self):
        """Opens the UDP transport socket and starts receiving traps"""
        raise NotImplementedError

    def close(self):
        """Closes the running SNMP engine and its associated ports"""
        raise NotImplementedError

    def auto_subscribe_observers(self):
        """Automatically subscribes all loaded TrapObserver subclasses to this trap receiver"""
        for observer_class in TrapObserver.__subclasses__():
            if not observer_class.WANTED_TRAPS:
                continue
            if observer_class in self._auto_subscribed_observers:
                continue
            else:
                self._auto_subscribed_observers.add(observer_class)
            observer_instance = observer_class(state=self.state, polldevs=self.polldevs, loop=self.loop)
            self.observe(observer_instance, *observer_instance.WANTED_TRAPS)

    def observe(self, subscriber: TrapObserver, *trap_types: List[TrapType]):
        """Adds a trap subscriber to the receiver"""
        for trap_type in trap_types:
            observers = self._observers.setdefault(trap_type, [])
            observers.append(subscriber)

    def get_observers_for(self, trap_type: TrapType) -> List[TrapObserver]:
        """Returns a list of trap observers for a given trap type"""
        return self._observers.get(trap_type, [])

    def add_community(self, community: str):
        """Adds a new community string that will be accepted on incoming packets"""
        if community in self._communities:
            return
        self._communities.add(community)

    async def dispatch_trap(self, trap: TrapMessage):
        """Dispatches incoming trap messages according to internal subscriptions"""
        observers = self.get_observers_for((trap.mib, trap.name))
        if not observers:
            _logger.debug("unknown trap: %s", trap)
            return

        for observer in observers:
            try:
                if not await observer.handle_trap(trap):
                    return
            except Exception:  # noqa
                _logger.exception("Unhandled exception in trap observer %r", observer)

    def _lookup_device(self, address: IPAddress) -> Optional[DeviceState]:
        """Looks up a device from Zino's running state from an IP address"""
        name = self.state.addresses.get(address)
        if name in self.state.devices:
            return self.state.devices[name]



#!/usr/bin/env python3
"""Fetch SNMP agent uptime using Zino high-level APIs"""

import argparse
import asyncio
import logging

from zino.config.polldevs import read_polldevs
from zino.snmp import import_snmp_backend
from zino.state import config

_log = logging.getLogger(__name__)


def main():
    args = parse_args()
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s - %(levelname)s - %(name)s (%(threadName)s) - %(message)s"
    )
    asyncio.run(run(args))


async def run(args: argparse.Namespace):
    devices, _ = read_polldevs(config.polling.file)
    device = devices[args.router]

    backend = import_snmp_backend(config.snmp.backend)
    backend.init_backend()
    snmp = backend.SNMP(device)
    response = await snmp.get("SNMPv2-MIB", "sysUpTime", 0)
    _log.info("Response from %s: %r", device.name, response.value)


def parse_args():
    devices, _ = read_polldevs(config.polling.file)
    parser = argparse.ArgumentParser(description="Fetch sysUptime from a device in the pollfile")
    parser.add_argument("router", type=str, help="Zino router name", choices=devices.keys())
    return parser.parse_args()


if __name__ == "__main__":
    main()



import argparse
import logging
from datetime import datetime, timezone
from typing import get_args

from zino.state import ZinoState
from zino.stateconverter.bfd_converter import set_bfd_state
from zino.stateconverter.bgp_converter import set_bgp_state
from zino.stateconverter.event_converter import set_event_state
from zino.stateconverter.flapping_converter import set_flapping_state
from zino.stateconverter.linedata import LineData
from zino.stateconverter.pm_converter import set_pm_state
from zino.stateconverter.port_converter import set_port_state
from zino.stateconverter.utils import load_state_to_dict, parse_ip
from zino.statemodels import CISCO_ENTERPRISE_ID, JUNIPER_ENTERPRISE_ID, AlarmType

_log = logging.getLogger(__name__)
_ignored_router_addrs: set[str] = set()


def create_state(old_state_file: str) -> ZinoState:
    new_state = ZinoState()
    # The returned dictionary defaults to empty list, so we dont need to check if the key exists
    # or manually set default value to avoid crashing if a variable is missing in the state file
    old_state = load_state_to_dict(old_state_file)

    set_bfd_state(old_state, new_state)
    set_bgp_state(old_state, new_state)
    set_event_state(old_state, new_state)
    set_port_state(old_state, new_state)
    set_flapping_state(old_state, new_state)
    set_pm_state(old_state, new_state)

    # Set vendor state
    for linedata in old_state["::isJuniper"]:
        set_is_juniper(linedata, new_state)
    for linedata in old_state["::isCisco"]:
        set_is_cisco(linedata, new_state)

    # Load device alarm state
    for linedata in old_state["::JNXalarms"]:
        set_jnx_alarms(linedata, new_state)

    # Load device boot time
    for linedata in old_state["::BootTime"]:
        set_boot_time(linedata, new_state)

    # Load address to router mapping
    for linedata in old_state["::AddrToRouter"]:
        set_addr_to_router(linedata, new_state)
    if _ignored_router_addrs:
        _log.error("Ignored invalid router addresses when converting AddrToRouter: %r", _ignored_router_addrs)

    # Load timestamps for when events were closed
    for linedata in old_state["::EventCloseTimes"]:
        set_event_close_times(linedata, new_state)

    return new_state


def set_is_cisco(linedata: LineData, state: ZinoState):
    is_cisco = bool(int(linedata.value))
    device = state.devices.get(linedata.identifiers[0])
    if is_cisco:
        device.enterprise_id = CISCO_ENTERPRISE_ID


def set_is_juniper(linedata: LineData, state: ZinoState):
    is_juniper = bool(int(linedata.value))
    device = state.devices.get(linedata.identifiers[0])
    if is_juniper:
        device.enterprise_id = JUNIPER_ENTERPRISE_ID


def set_jnx_alarms(linedata: LineData, state: ZinoState):
    device = state.devices.get(linedata.identifiers[0])
    alarm_type = linedata.identifiers[1]
    assert alarm_type in get_args(AlarmType)
    alarm_count = int(linedata.value)
    if not device.alarms:
        device.alarms = {}
    device.alarms[alarm_type] = alarm_count


def set_boot_time(linedata: LineData, state: ZinoState):
    device = state.devices.get(linedata.identifiers[0])
    timestamp = int(linedata.value)
    device.boot_time = datetime.fromtimestamp(timestamp, tz=timezone.utc)


def set_addr_to_router(linedata: LineData, state: ZinoState):
    """Corresponds to ZinoState.addresses"""
    ip_string = linedata.identifiers[0]
    try:
        ip = parse_ip(ip_string)
    except ValueError:
        _ignored_router_addrs.add(ip_string)
        return
    device_name = linedata.value
    state.addresses[ip] = device_name


def set_event_close_times(linedata: LineData, state: ZinoState):
    """Records when event was closed"""
    event_id = int(linedata.identifiers[0])
    event = state.events.events.get(event_id)
    if not event:
        _log.error("EventCloseTimes refers to non existing event %d", event_id)
        return
    close_time = datetime.fromtimestamp(int(linedata.value), tz=timezone.utc)
    event.closed = close_time


def get_parser():
    parser = argparse.ArgumentParser(description="Convert Zino1 state to Zino2 compatible state")
    parser.add_argument(
        "input",
        help="Absolute path to the Zino1 state you want to convert",
    )
    parser.add_argument(
        "output",
        help="Absolute path to where the new Zino2 state should be dumped",
    )
    return parser


def main():
    logging.basicConfig(format="%(message)s", level=logging.INFO)
    parser = get_parser()
    args = parser.parse_args()
    state = create_state(args.input)
    state.dump_state_to_file(args.output)


if __name__ == "__main__":
    main()



from zino.stateconverter.linedata import get_line_data


def test_only_quotes_on_start_and_end_of_value_are_removed():
    test_line = 'set ::EventAttrs_100(log) "{15000000 intf "ge-0/0/1" is down}"'
    expected_value = '{15000000 intf "ge-0/0/1" is down}'
    linedata = get_line_data(test_line)
    assert linedata.value == expected_value


def test_identifiers_are_parsed_correctly_for_event_attrs():
    test_line = 'set ::EventAttrs_100(log) "{15000000 intf "ge-0/0/1" is down}"'
    expected_identifiers = tuple(["log", "100"])
    linedata = get_line_data(test_line)
    assert linedata.identifiers == expected_identifiers


def test_identifiers_are_parsed_in_correct_order():
    test_line = 'set ::SomeCommand(id1,id2,id3) "some value"'
    expected_identifiers = tuple(["id1", "id2", "id3"])
    linedata = get_line_data(test_line)
    assert linedata.identifiers == expected_identifiers


def test_can_parse_global_setters():
    test_line = 'set ::pm::lastid "10"'
    linedata = get_line_data(test_line)
    assert linedata.value == "10"
    assert linedata.identifiers is None



from zino.api.server import ZinoServer
from zino.state import ZinoState


def test_zinoserver_should_serve_without_error(event_loop):
    server = ZinoServer(loop=event_loop, state=ZinoState(), polldevs=dict())
    server.serve()
    server.notify_server.close()
    server.api_server.close()
